#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 16 11:53:56 2019

@author: psakicki
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import dateutil
import numpy as np
import os 
import shutil
import string
import subprocess
import re

#### geodeZYX modules
from geodezyx import conv
from geodezyx import utils

##########  END IMPORT  ##########

#  _____  _____ _   _ ________   __   _____       _ _ _
# |  __ \|_   _| \ | |  ____\ \ / /  / ____|     | (_) |
# | |__) | | | |  \| | |__   \ V /  | (___  _ __ | |_| |_ ___ _ __
# |  _  /  | | | . ` |  __|   > <    \___ \| '_ \| | | __/ _ \ '__|
# | | \ \ _| |_| |\  | |____ / . \   ____) | |_) | | | ||  __/ |
# |_|  \_\_____|_| \_|______/_/ \_\ |_____/| .__/|_|_|\__\___|_|
#                                          | |
#                                          |_|

def check_if_compressed_rinex(rinex_path):
    boolout = bool(re.search('.*((d|o)\.(Z)|(gz))$', rinex_path))
    return boolout

def crz2rnx(rinex_path,outdir='',force=True,path_of_crz2rnx='CRZ2RNX'):
    """ assuming that CRZ2RNX is in the system PATH per default """

    if not os.path.isfile(rinex_path):
        raise Exception( rinex_path + ' dont exists !')

    if force:
        forcearg = '-f'
    else:
        forcearg = ''

    curdir = os.getcwd()
    command = path_of_crz2rnx + ' -c ' + forcearg + ' ' + rinex_path
    if outdir == '':
        outdir = os.path.dirname(rinex_path)

    out_rinex_name_splited = os.path.basename(rinex_path).split('.')
    out_rinex_name = '.'.join(out_rinex_name_splited[:-1])
    out_rinex_name = out_rinex_name[:-1] + 'o'
    out_rinex_path = os.path.join(outdir,out_rinex_name)

    if os.path.isfile(out_rinex_path) and not force:
        print("INFO : crz2rnx : ", out_rinex_path , 'already exists, skiping ...')
    else:
        os.chdir(outdir)
        #stream = os.popen(command)

        proc = subprocess.Popen(command, shell=True,
                            stdout=subprocess.PIPE,executable='/bin/bash')
        status = proc.wait()

        if status != 0 or not os.path.isfile(out_rinex_path):
            print('ERR : crz2rnx : ',out_rinex_path,"not created ! :(")
        else:
            print('INFO : crz2rnx : ',out_rinex_path, 'created :)')
        #    print stream.read()
        os.chdir(curdir)
    return out_rinex_path



def crz2rnx_bad(crinex_in_path,outdir='',force=True,path_of_crz2rnx='CRZ2RNX'):
    """ assuming that CRZ2RNX is in the system PATH per default """

    if not os.path.isfile(crinex_in_path):
        raise Exception( crinex_in_path  + ' dont exists !')

    if force:
        forcearg = '-f'
    else:
        forcearg = ''

    command = path_of_crz2rnx + ' ' + forcearg + ' ' + crinex_in_path
    convert_dir = os.path.dirname(crinex_in_path)

    if outdir == '':
        outdir = convert_dir

    out_rinex_name_splited = os.path.basename(crinex_in_path).split('.')
    out_rinex_name = '.'.join(out_rinex_name_splited[:-1])
    out_rinex_name = out_rinex_name[:-1] + 'o'

    out_rinex_path  = os.path.join(outdir,out_rinex_name)
    temp_rinex_path = os.path.join(convert_dir,out_rinex_name)

    if os.path.isfile(out_rinex_path) and not force:
        print("INFO : crz2rnx : ", out_rinex_path , 'already exists, skiping ...')
    else:
        stream = os.popen(command)
        print('output in',outdir)
        try:
            shutil.move(temp_rinex_path, outdir)
        except:
            pass

        #    print stream.read()
#    os.chdir(curdir)
    return out_rinex_path

def rnx2crz(rinex_path,outdir='',force=True,path_of_rnx2crz='RNX2CRZ'):
    """ assuming that RNX2CRZ is in the system PATH per default """

    if not os.path.isfile(rinex_path):
        raise Exception( rinex_path + ' dont exists !')

    if force:
        forcearg = '-f'
    else:
        forcearg = ''

    curdir = os.curdir
    command = path_of_rnx2crz + ' ' + forcearg + ' ' + rinex_path
    if outdir == '':
        outdir = os.path.dirname(rinex_path)

    out_rinex_name_o = os.path.basename(rinex_path)
    out_rinex_name_d_Z = out_rinex_name_o[:-1] + 'd.Z'
    out_rinex_path = os.path.join(outdir,out_rinex_name_d_Z)

    if os.path.isfile(out_rinex_path) and not force:
        print("INFO : rnx2crz : ", out_rinex_path , 'already exists, skiping ...')
    else:
        os.chdir(outdir)
        stream = os.popen(command)
        print(command,', output in',outdir)
        #    print stream.read()
    os.chdir(curdir)
    return out_rinex_path

def rinex_regex(compressed=True,compiled=False):
    if compressed:
        regexstr = "....[0-9]{3}.\.[0-9]{2}((d\.(Z)|(gz))|(o)|(d))"
    else:
        regexstr = "....[0-9]{3}.\.[0-9]{2}o"

    if compiled:
        return re.compile(regexstr)
    else:
        return regexstr


def rinex_regex_new_name(compressed=True,compiled=False):
    if compressed:
        regexstr = ".{4}[0-9]{2}.{3}_(R|S|U)_[0-9]{11}_[0-9]{2}\w_[0-9]{2}\w_\w{2}\.\w{3}\.gz"
    else:
        regexstr = ".{4}[0-9]{2}.{3}_(R|S|U)_[0-9]{11}_[0-9]{2}\w_[0-9]{2}\w_\w{2}\.\w{3}"

    if compiled:
        return re.compile(regexstr)
    else:
        return regexstr


def rinex_read_epoch(input_rinex_path_or_string,interval_out=False,
                    add_tzinfo=False,out_array=True):
    """
    input_rinex_path_or_string :

        can be the path of a RINEX or directly the RINEX content as a string

    161019 : dirty copier coller de rinex start end
    """
    epochs_list = []
    rinex_60sec = False



    if  utils.is_iterable(input_rinex_path_or_string):
        Finp = input_rinex_path_or_string
    elif os.path.isfile(input_rinex_path_or_string):
        Finp = open(input_rinex_path_or_string)
    else:
        Finp = input_rinex_path_or_string


    for line in Finp:
        epoch_rnx2=re.search('^ {1,2}([0-9]{1,2} * ){5}',line)
        
        if epoch_rnx2:
            fraw = line.split() #fraw est la pour gerer les fracion de sec ...
            fraw = fraw[0:6]
            fraw = [float(e) for e in fraw]
            f = [int(e) for e in fraw]
            msec = (fraw[5] - np.floor(fraw[5]))
            msec = np.round(msec,4)
            msec = int(msec * 10**6)
            f.append( msec )  # ajour des fractions de sec

            if f[0] < 50:
                f[0] = f[0] + 2000
            else:
                f[0] = f[0] + 1900
            if f[5] == 60: # cas particulier rencontré dans des rinex avec T = 60sec
                rinex_60sec = True
                f[5] = 59
            else:
                rinex_60sec = False
            try:
                epochdt = dt.datetime(*f)
                if rinex_60sec:
                    epochdt = epochdt + dt.timedelta(seconds=1)
                epochs_list.append(epochdt)

            except:
                print("ERR : rinex_read_epoch : " , f)
                
    if add_tzinfo:
        epochs_list = [ e.replace(tzinfo=dateutil.tz.tzutc()) for e in epochs_list]

    if out_array:
        return np.array(epochs_list)
    else:
        return epochs_list

def same_day_rinex_check(rinex1,rinex2):
    if os.path.basename(rinex1)[4:7] == os.path.basename(rinex2)[4:7]:
        return True
    else:
        return False


def rinex_start_end(input_rinex_path,interval_out=False,
                    add_tzinfo=False,verbose = True,
                    safety_mode = True):
    """
    safety_mode :

        if the epoch reading fails (e.g. in case of a compressed RINEX)
        activate a reading of the header and the file name as backup.


    une liste d'epochs en début et fin de fichier
    => en trouver le min et le max
    NB : FAIRE UN FONCTION READ EPOCH A L'OCCAZ
    NBsuite : c'est fait au 161018 mais par contre c'est un dirty copier coller
    """
    epochs_list = []
    Head = utils.head(input_rinex_path,1500)
    epochs_list_head = rinex_read_epoch(Head,interval_out=interval_out,
                                        add_tzinfo=add_tzinfo,out_array=False)


    Tail =  utils.tail(input_rinex_path,1500)
    epochs_list_tail = rinex_read_epoch(Tail,interval_out=interval_out,
                                        add_tzinfo=add_tzinfo,out_array=False)

    epochs_list = epochs_list_head + epochs_list_tail

    if len(epochs_list) == 0:
        first_epoch = conv.rinexname2dt(input_rinex_path)
        alphabet = list(string.ascii_lowercase)

        if os.path.basename(input_rinex_path)[7] in alphabet:
            last_epoch = first_epoch + dt.timedelta(hours=1)
        else:
            last_epoch = first_epoch + dt.timedelta(hours=24,seconds=-1)
    else:
        first_epoch = np.min(epochs_list)
        last_epoch = np.max(epochs_list)

    if add_tzinfo:
        first_epoch = first_epoch.replace(tzinfo=dateutil.tz.tzutc())
        last_epoch = last_epoch.replace(tzinfo=dateutil.tz.tzutc())

    if verbose:
        print("first & last epochs : " , first_epoch , last_epoch)
    if not interval_out:
        return first_epoch , last_epoch
    else:
        interv_lis = np.diff(epochs_list)
        interv_lis = [e.seconds + e.microseconds * 10**-6 for e in interv_lis]
        interval   = utils.most_common(interv_lis)
        print("interval : " , interval , last_epoch)

        #return interv_lis , epochs_list
        return first_epoch , last_epoch , interval

def rinex_session_id(first_epoch,last_epoch,full_mode=False):
    """
    full_mode:
        gives the letter of the starting session  & the length in hour
        of the session
    """

    alphabet = list(string.ascii_lowercase)
    rinex_length = last_epoch - first_epoch
    if rinex_length.seconds > 86000:
        rnx_interval_ext = '0'
    else:
        if not full_mode:
            rnx_interval_ext = alphabet[first_epoch.hour]
        else:
            rnx_interval_ext = alphabet[first_epoch.hour]  \
                               + str(int(rinex_length.seconds / 3600.))

    return rnx_interval_ext

#def rinex_spliter(input_rinex_path,output_directory,stat_out_name='',
#                  interval_size=24,compress=False,shift = 0):
#    """
#    if shift != 0:
#    the start/end of a session is shifted of shift minutes
#    """
#
#    if stat_out_name == '':
#        stat_out_name = os.path.basename(input_rinex_path)[0:4]
#
#    # check if the RINEX is compressed ...
#    bool_comp_rnx = check_if_compressed_rinex(input_rinex_path)
#    # ... if not crz2rnx !
#    if bool_comp_rnx:
#        input_rinex_path = crz2rnx(input_rinex_path)
#
#    inp_rinex_obj=open(input_rinex_path,'r+')
#    out_dir = output_directory # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable
#
#    if not os.path.exists(out_dir):
#        os.makedirs(out_dir)
#    os.chdir(out_dir)
#
#    first_epoch , last_epoch = rinex_start_end(input_rinex_path)
#
#    # In this function, Date are truncated epochs
#    # (only the day if interval == 24 , + the hour else)
#    first_date = datetime.datetime(first_epoch.year,first_epoch.month,
#                                   first_epoch.day,first_epoch.hour)
#    last_date = datetime.datetime(last_epoch.year,last_epoch.month,
#                                  last_epoch.day,last_epoch.hour)+datetime.timedelta(hours=1)
#
#    print "first & last dates (truncated) : " , first_date , last_date
#
#    time_interval = genefun.get_interval(first_date,last_date,
#                                         datetime.timedelta(hours=interval_size))
#
#    alphabet = list(string.ascii_lowercase)
#
#    rinex_out_name_lis = []
#
#    for i,curr_date in enumerate(time_interval):
#        if interval_size == 24:
#            rnx_interval_ext = '0.'
#        else:
#            rnx_interval_ext = alphabet[curr_date.hour] + '.'
#
#        p = subprocess.Popen('',executable='/bin/bash', stdin=subprocess.PIPE , stdout=subprocess.PIPE , stderr=subprocess.PIPE)
#        command = 'teqc ' + '-O.mo ' + stat_out_name.upper() +' -st ' +  curr_date.strftime('%y%m%d%H%M%S') + ' +dh ' + str(interval_size) + ' ' + input_rinex_path
#        print command
#        rinex_out_name = stat_out_name + curr_date.strftime('%j') + rnx_interval_ext + curr_date.strftime('%y') + 'o'
#        err_log_name = curr_date.strftime('%j') + rnx_interval_ext + "err.log"
#        print rinex_out_name
#        stdout,stderr = p.communicate( command )
#        std_file = open(rinex_out_name, "w")
#        std_file.write(stdout)
#        std_file.close()
#        if stderr != '':
#            print err_log_name + " is not empty, must be checked !"
#            print stderr
#            err_file = open(err_log_name, "w")
#            err_file.write(stderr)
#            err_file.close()
#
#        rnx_splt_path = os.path.join(out_dir,rinex_out_name)
#        if compress:
#            rinex_out_final = rnx2crz(rnx_splt_path)
#            os.remove(rnx_splt_path)
#        else:
#            rinex_out_final = rnx_splt_path
#
#        rinex_out_name_lis.append(rinex_out_final)
#
#    return rinex_out_name_lis


def rinex_spliter(input_rinex_path,output_directory,stat_out_name='',
                  interval_size=24,compress=False,shift = 0,
                  inclusive = False,teqc_cmd='teqc'):
    """
    if shift != 0:
    the start/end of a session is shifted of shift minutes

    inclusive:
    delta of exaclty interval_size => add the 1st epoch of the next sess
    not inclusive:
    delta of interval_size - 1s
    """
    """
    Split RINEX file using teqc

    Parameters
    ----------
    input_rinex_path : str
        Path of the input RINEX

    output_directory : str
        Description param2
        
    stat_out_name : str
        Station name for the output RINEX
        
    interval_size : int
        Size of the splitted interval
    
    compress : bool
        compress the outputed RINEX    

    interval_size : int
        Size of the splitted interval 
        
        
    FINIR CE HEADER !!!!
                
    Returns
    -------
    out1 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description out1
    
    out2 : float or int or str or dict or n-tuple or bool or list or numpy.array
        Description out2
        
    Note
    ----
    Misc. Notes

    Source
    ------
    www.forum-source.com
    
    Examples
    --------
    >> answer
    42    
    """

    if stat_out_name == '':
        stat_out_name = os.path.basename(input_rinex_path)[0:4]

    # check if the RINEX is compressed ...
    bool_comp_rnx = check_if_compressed_rinex(input_rinex_path)

    # ... if not crz2rnx !
    if bool_comp_rnx:
        input_rinex_path = crz2rnx(input_rinex_path)

    inp_rinex_obj=open(input_rinex_path,'r+')
    out_dir = output_directory # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    os.chdir(out_dir)

    first_epoch , last_epoch = rinex_start_end(input_rinex_path)

    # In this function, Date are truncated epochs
    # (only the day if interval == 24 , + the hour else)
    first_date = dt.datetime(first_epoch.year,first_epoch.month,
                                   first_epoch.day,first_epoch.hour)
    last_date = dt.datetime(last_epoch.year,last_epoch.month,
                                  last_epoch.day,last_epoch.hour)+dt.timedelta(hours=1)

    print("first & last dates (truncated) : " , first_date , last_date)

    time_interval = utils.get_interval(first_date,last_date,
                                         dt.timedelta(hours=interval_size))

    alphabet = list(string.ascii_lowercase)

    rinex_out_name_lis = []

    for i,curr_date in enumerate(time_interval):

        if shift != 0:
            print('WARN : rinex_spliter : shifted mode on, be careful !!!')


        if bool(shift) and i != 0:
            curr_date = curr_date + dt.timedelta(minutes = shift)
            interval_size_ope = interval_size
        elif bool(shift) and i == 0 :
            interval_size_ope = interval_size + float(shift) / 60.
        else:
            interval_size_ope = interval_size

        if not inclusive:
            interval_size_ope = interval_size_ope - 1/3600.

        if not bool(shift):
            if interval_size == 24:
                rnx_interval_ext = '0.'
            else:
                rnx_interval_ext = alphabet[curr_date.hour] + '.'
        else:
            if interval_size == 24:
                rnx_interval_ext = '0.'
            else:
                rnx_interval_ext = alphabet[curr_date.hour]  +'.'

        p = subprocess.Popen('',executable='/bin/bash', stdin=subprocess.PIPE , stdout=subprocess.PIPE , stderr=subprocess.PIPE)
        # - 1/3600. # one_sec2h
        command = teqc_cmd  + ' -O.mo ' + stat_out_name.upper() +' -st ' +  curr_date.strftime('%y%m%d%H%M%S') + ' +dh ' + str(interval_size_ope) + ' ' + input_rinex_path
        print(command)
        rinex_out_name = stat_out_name + curr_date.strftime('%j') + rnx_interval_ext + curr_date.strftime('%y') + 'o'
        err_log_name = curr_date.strftime('%j') + rnx_interval_ext + "err.log"
        print(rinex_out_name)
        stdout,stderr = p.communicate( command.encode() )
        std_file = open(rinex_out_name, "w")
        std_file.write(stdout.decode("utf-8"))
        std_file.close()
        if stderr:
            print(err_log_name + " is not empty, must be checked !")
            print(stderr.decode("utf-8"))
            err_file = open(err_log_name, "w")
            err_file.write(stderr.decode("utf-8"))
            err_file.close()

        rnx_splt_path = os.path.join(out_dir,rinex_out_name)
        if compress:
            rinex_out_final = rnx2crz(rnx_splt_path)
            os.remove(rnx_splt_path)
        else:
            rinex_out_final = rnx_splt_path

        rinex_out_name_lis.append(rinex_out_final)

    return rinex_out_name_lis



def teqc_qc(rinex_path,quick_mode=False,optional_args=''):
    """
    quick mode : reduced qc and no summary file written
    """
    if quick_mode:
        qcmode = "+qcq"
    else:
        qcmode = "+qc"

    # check if the RINEX is compressed ...
    bool_comp_rnx = check_if_compressed_rinex(rinex_path)

    # ... if not crz2rnx !
    if bool_comp_rnx:
        rinex_path_work = crz2rnx(rinex_path)
    else:
        rinex_path_work = rinex_path



    kommand = " ".join(("teqc" , qcmode , optional_args , rinex_path))

    print(kommand)

    proc = subprocess.Popen(kommand, shell=True,
                            stdout=subprocess.PIPE,executable='/bin/bash')
    status = proc.wait()

    if bool_comp_rnx:
        os.remove(rinex_path_work)

    if status:
        print("ERR : teqc qc : crash for " + rinex_path + ", code " + str(proc.poll()))
        return None

    output = proc.stdout.read()

    return output.decode('ASCII')

def rinex_renamer(input_rinex_path,output_directory,stat_out_name='',remove=False):

    if stat_out_name == '':
        stat_out_name = os.path.basename(input_rinex_path)[0:4]

    stat_out_name = stat_out_name.lower()

    inp_rinex_obj=open(input_rinex_path,'r+')
    out_dir = output_directory # nom redondant mais j'ai la flemme d'aller corriger le nom de la variable

    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    os.chdir(out_dir)

    first_epoch , last_epoch = rinex_start_end(input_rinex_path)
    rnx_interval_ext = rinex_session_id(first_epoch,last_epoch) + '.'
    rinex_out_name = stat_out_name + first_epoch.strftime('%j') + rnx_interval_ext + first_epoch.strftime('%y') + 'o'

    print(rinex_out_name)

    output_rinex_path = os.path.join(out_dir,rinex_out_name)

    if input_rinex_path != output_rinex_path:
        print("INFO : copy of ", input_rinex_path , ' to ' , output_rinex_path)
        shutil.copy(input_rinex_path,output_rinex_path)

        if remove and os.isfile(output_rinex_path) :
            print("INFO : removing " , input_rinex_path)
            os.remove(input_rinex_path)
    else:
        print("INFO : " , input_rinex_path)
        print("and", output_rinex_path ,"are the same file")
        print("nothing's done ...")

    return output_rinex_path


# def rinex_renamer_gfz_odc_2_igs(rinex_in,output_dir=None):
    
    
    
    
    



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: objmaterial.py,v 1.7 2005/08/28 19:42:43 mbaas Exp $

## \file objmaterial.py
## Contains the OBJMaterial class.

import os.path, sys
from . import protocols
from .Interfaces import *
from .slots import *
from .material import Material
from . import ribexport
from . import _Image as Image
from .expression import Expression
from . import _core
from .cgtypes import *

# OBJTextureMap
class OBJTextureMap:
    """A texture map definition as it appears in OBJ/MTL files.
    """

    def __init__(self,
                 filename,
                 offset = (0,0,0),
                 scale = (1,1,1),
                 turb = (0,0,0),
                 mm = (0.0, 1.0),
                 clamp = False,
                 blendu = True,
                 blendv = True,
                 bumpsize = None,
                 refltype = None):
        
        self.filename = filename
        self.offset = offset
        self.scale = scale
        self.turb = turb
        self.mm = mm
        self.clamp = clamp
        self.blendu = blendu
        self.blendv = blendv
        if bumpsize!=None:
            self.bumpsize = bumpsize
        if refltype!=None:
            self.refltype = refltype
        
        
# OBJMaterial
#class OBJMaterial(Material):
class OBJMaterial(_core.GLMaterial):
    """This class represents a material as it appears in OBJ (or rather MTL) files.

    """

    protocols.advise(instancesProvide=[ribexport.IMaterial])
    
    def __init__(self,
                 name="OBJMaterial",
                 illum = 2,
                 Ka = (0.2, 0.2, 0.2),
                 Kd = (0.8, 0.8, 0.8),
                 Ks = (0.0, 0.0, 0.0),
                 Ke = (0.0, 0.0, 0.0),
                 Ns = 0.0,
                 Ni = 1.0,
                 d = 1.0,
                 Tr = 1.0,
                 Tf = (1.0, 1.0, 1.0),
                 sharpness = 0.0,
                 map_Ka = None,
                 map_Kd = None,
                 map_Ks = None,
                 map_Ke = None,
                 map_Ns = None,
                 map_d = None,
                 map_Bump = None,
                 refl = None,  # one single map or a list of maps
                 density = 1.0,
                 ):

#        Material.__init__(self, name=name, density=density)
        _core.GLMaterial.__init__(self, name, density)

        # Create a Kd slot and connect it with the GL diffuse slot
        self.Kd_slot = Vec3Slot(vec3(Kd))
        self.addSlot("Kd", self.Kd_slot)
        self.diffuse_adapter = Expression("(col.x,col.y,col.z,1)", col=vec3())
        self.Kd_slot.connect(self.diffuse_adapter.col_slot)
        self.diffuse_adapter.output_slot.connect(self.diffuse_slot)

        self.illum = illum
        self.Ka = Ka
#        self.Kd = Kd
        self.Ks = Ks
        self.Ke = Ke
        self.Ns = Ns
        self.Ni = Ni
        self.d = d
        self.Tr = Tr
        self.Tf = vec3(Tf)
        self.sharpness = sharpness
        self.map_Ka = map_Ka
        self.map_Kd = map_Kd
        self.map_Ks = map_Ks
        self.map_Ke = map_Ke
        self.map_Ns = map_Ns
        self.map_d = map_d
        self.map_Bump = map_Bump

        if refl==None:
            refl = []
        else:
            # Turn refl into a list
            try:
                refl = list(refl)
            except:
                refl = [refl]
        self.refl = refl

        # "Hide" the GLMaterial slots
        self.removeSlot("ambient")
        self.removeSlot("diffuse")
        self.removeSlot("specular")
        self.removeSlot("shininess")
        self.removeSlot("emission")

    exec(slotPropertyCode("Kd"))

    def mtlDefinition(self):
        s = """illum %d
Ka %f %f %f
Kd %f %f %f
Ks %f %f %f
Ke %f %f %f
Ns %f
Ni %f
"""%(self.illum,
     self.Ka[0], self.Ka[1], self.Ka[2],
     self.Kd[0], self.Kd[1], self.Kd[2],
     self.Ks[0], self.Ks[1], self.Ks[2],
     self.Ke[0], self.Ke[1], self.Ke[2],
     self.Ns,
     self.Ni)
        if self.d!=1.0:
            s += "d %f\n"%self.d
        if self.Tr!=1.0:
            s += "Tr %f\n"%self.d
        if self.Tf[0]!=1.0 or self.Tf[1]!=1.0 or self.Tf[2]!=1.0:
            s += "Tf %f %f %f\n"%(self.Tf[0], self.Tf[1], self.Tf[2])
        s += self.mapDefinition("map_Ka", self.map_Ka)
        s += self.mapDefinition("map_Kd", self.map_Kd)
        s += self.mapDefinition("map_Ks", self.map_Ks)
        s += self.mapDefinition("map_Ke", self.map_Ke)
        s += self.mapDefinition("map_Ns", self.map_Ns)
        s += self.mapDefinition("map_d", self.map_d)
        s += self.mapDefinition("map_Bump", self.map_Bump)
        for map in self.refl:
            s += self.mapDefinition("refl", map)
        return s

    # mapDefinition
    def mapDefinition(self, mapname, map):
        """Return the map defintion line or an empty string.
        """
        if map==None:
            return ""

        onoff = {True:"on", False:"off"}

        res = mapname
        # -o
        u,v,w = map.offset
        if u!=0 or v!=0 or w!=0:
            res += " -o %f %f %f"%(u,v,w)
        # -s
        u,v,w = map.scale
        if u!=1 or v!=1 or w!=1:
            res += " -s %f %f %f"%(u,v,w)
        # -t
        u,v,w = map.turb
        if u!=0 or v!=0 or w!=0:
            res += " -t %f %f %f"%(u,v,w)
        # -mm
        base, gain = map.mm
        if base!=0 or gain!=1:
            res += " -mm %f %f"%(base, gain)
        # -clamp
        if map.clamp:
            res += " -clamp %s"%onoff[map.clamp]
        # -blendu
        if not map.blendu:
            res += " -blendu %s"%onoff[map.blendu]
        # -blendv
        if not map.blendv:
            res += " -blendv %s"%onoff[map.blendv]
        # -bm
        if mapname=="map_Bump" and getattr(map, "bumpsize", 1)!=1:
            res += " -bm %f"%map.bumpsize
        # -type
        if mapname=="refl":
            res += " -type %s"%getattr(map, "refltype", "sphere")

        return res+" %s\n"%map.filename        
        
        
    def createPasses(self):
        """Returns a list of RenderPass objects."""
        texdefs = []
        if self.map_Ka!=None:
            texdefs.append((self.map_Ka.filename, "periodic", "periodic", "gaussian", 1.0, 1.0, {}))
        if self.map_Kd!=None:
            texdefs.append((self.map_Kd.filename, "periodic", "periodic", "gaussian", 1.0, 1.0, {}))
        if self.map_Ks!=None:
            texdefs.append((self.map_Ks.filename, "periodic", "periodic", "gaussian", 1.0, 1.0, {}))
        if self.map_d!=None:
            texdefs.append((self.map_d.filename, "periodic", "periodic", "gaussian", 1.0, 1.0, {}))
        if self.map_Bump!=None:
            texdefs.append((self.map_Bump.filename, "periodic", "periodic", "gaussian", 1.0, 1.0, {}))
        
        if texdefs!=[]:
            return [ribexport.TexPass(maps=texdefs)]
        else:
            return []

    def preProcess(self, exporter):
        """Preprocessing method.

        This method is called before the image is rendered and can be used
        to create or copy image maps.
        """
        pass
            

    def color(self):
        """Return the color for the RiColor() call or None.
        """
        return None
#        return self.Kd

    def opacity(self):
        """Return the opacity for the RiOpacity() call or None.
        """
        return None
#        a = self.Tr*self.d
#        return (a,a,a)

    def surfaceShaderName(self):
        """Returns the name of the corresponding surface shader or None.
        """
        return "matobj"

    def surfaceShaderSource(self):
        """Returns surface shader source code as a string or None.

        If the return value is None, then shaderName() must return
        the name of the shader to use.
        """
        return """// OBJ/MTL material shader
        
surface $SHADERNAME(float illum = 2;
           color Ka = color "rgb" (0.2, 0.2, 0.2);
           color Kd = color "rgb" (0.8, 0.8, 0.8);
           color Ks = color "rgb" (0.8, 0.8, 0.8);
           color Ke = color "rgb" (0, 0, 0);
           float Ni = 0.0;
           float Ns = 0.0;
           float d = 1.0;
           float Tr = 1.0;
           color Tf = color "rgb" (1.0, 1.0, 1.0);
           
           string map_Ka = "";
           float map_Ka_offset[3] = {0, 0, 0};
           float map_Ka_scale[3] = {1, 1, 1};
           
           string map_Kd = "";
           float map_Kd_offset[3] = {0, 0, 0};
           float map_Kd_scale[3] = {1, 1, 1};
           
           string map_Ks = "";
           float map_Ks_offset[3] = {0, 0, 0};
           float map_Ks_scale[3] = {1, 1, 1};

           string map_Ke = "";
           float map_Ke_offset[3] = {0, 0, 0};
           float map_Ke_scale[3] = {1, 1, 1};

           string map_Ns = "";
           float map_Ns_offset[3] = {0, 0, 0};
           float map_Ns_scale[3] = {1, 1, 1};

           string map_d = "";
           float map_d_offset[3] = {0, 0, 0};
           float map_d_scale[3] = {1, 1, 1};

           varying point Pref = point(0,0,0);
           )
{
  BAKE_BEGIN
  normal Nf = BAKE_NORMAL(N);
  color _Ka = Ka;
  color _Kd = Kd;
  color _Ks = Ks;
  color _Ke = Ke;
  float _Ns = Ns;
  float _d = d;

  if (map_Ka!="")
  {
    _Ka = texture(map_Ka,
                  s*map_Ka_scale[0]+map_Ka_offset[0],
                  (1-t)*map_Ka_scale[1]+map_Ka_offset[1]);
  }

  if (map_Kd!="")
  {
    _Kd = texture(map_Kd,
                  s*map_Kd_scale[0]+map_Kd_offset[0],
                  (1-t)*map_Kd_scale[1]+map_Kd_offset[1]);
  }

  if (map_Ks!="")
  {
    _Ks = texture(map_Ks,
                  s*map_Ks_scale[0]+map_Ks_offset[0],
                  (1-t)*map_Ks_scale[1]+map_Ks_offset[1]);
  }

  if (map_Ke!="")
  {
    _Ke = texture(map_Ke,
                  s*map_Ke_scale[0]+map_Ke_offset[0],
                  (1-t)*map_Ke_scale[1]+map_Ke_offset[1]);
  }

  if (map_Ns!="")
  {
    _Ns = texture(map_Ns,
                  s*map_Ns_scale[0]+map_Ns_offset[0],
                  (1-t)*map_Ns_scale[1]+map_Ns_offset[1]);
  }

  if (map_d!="")
  {
    _d = texture(map_d,
                 s*map_d_scale[0]+map_d_offset[0],
                 (1-t)*map_d_scale[1]+map_d_offset[1]);
  }

  if (illum==0)
  {
    Ci = _Kd;
  }
  else if (illum==1)
  {
    Ci = _Ka + _Kd*diffuse(Nf);
  }
  else
  {
    // Ns==0? then disable specular part
    if (_Ns<1E-6)
    {
      _Ks = color "rgb" (0,0,0);
    }
    Ci = _Ka*ambient() + _Kd*diffuse(Nf) + _Ks*specular(Nf,-normalize(I),1/_Ns) + _Ke;
  }  

  Oi = _d*Tr*Tf;
  Ci *= Oi;
  BAKE_END
}        
        """

    def surfaceShaderParams(self, passes):
        """Return a dictionary with shader parameters and their values."""
        res = {"uniform float illum" : self.illum,
               "uniform color Ka" : self.Ka,
               "uniform color Kd" : self.Kd,
               "uniform color Ks" : self.Ks,
               "uniform color Ke" : self.Ke,
               "uniform float Ns" : self.Ns,
               "uniform float Ni" : self.Ni,
               "uniform float d" : self.d,
               "uniform float Tr" : self.Tr,
               "uniform color Tf" : self.Tf
               }

        if self.map_Ka!=None:
            texname = os.path.basename(self.map_Ka.filename)
            name, ext = os.path.splitext(texname)
            res["uniform string map_Ka"] = name+".tex"       
            res["uniform float[3] map_Ka_offset"] = self.map_Ka.offset
            res["uniform float[3] map_Ka_scale"] = self.map_Ka.scale

        if self.map_Kd!=None:
            texname = os.path.basename(self.map_Kd.filename)
            name, ext = os.path.splitext(texname)
            res["uniform string map_Kd"] = name+".tex"       
            res["uniform float[3] map_Kd_offset"] = self.map_Kd.offset
            res["uniform float[3] map_Kd_scale"] = self.map_Kd.scale

        if self.map_Ks!=None:
            texname = os.path.basename(self.map_Ks.filename)
            name, ext = os.path.splitext(texname)
            res["uniform string map_Ks"] = name+".tex"       
            res["uniform float[3] map_Ks_offset"] = self.map_Ks.offset
            res["uniform float[3] map_Ks_scale"] = self.map_Ks.scale

        if self.map_Ke!=None:
            texname = os.path.basename(self.map_Ke.filename)
            name, ext = os.path.splitext(texname)
            res["uniform string map_Ke"] = name+".tex"  
            res["uniform float[3] map_Ke_offset"] = self.map_Ke.offset
            res["uniform float[3] map_Ke_scale"] = self.map_Ke.scale

        if self.map_Ns!=None:
            texname = os.path.basename(self.map_Ns.filename)
            name, ext = os.path.splitext(texname)
            res["uniform string map_Ns"] = name+".tex"
            res["uniform float[3] map_Ns_offset"] = self.map_Ns.offset
            res["uniform float[3] map_Ns_scale"] = self.map_Ns.scale

        if self.map_d!=None:
            texname = os.path.basename(self.map_d.filename)
            name, ext = os.path.splitext(texname)
            res["uniform string map_d"] = name+".tex"       
            res["uniform float[3] map_d_offset"] = self.map_d.offset
            res["uniform float[3] map_d_scale"] = self.map_d.scale

        return res
        
    def surfaceShaderTransform(self):
        return mat4(1)

    def displacementShaderName(self):
        if self.map_Bump==None:
            return None
        
        return "matobj_bump"

   
    def displacementShaderSource(self):
        if self.map_Bump==None:
            return None
        
        return """// OBJ/MTL material shader (bump)
        
displacement $SHADERNAME(
           string map_Bump = "";
           float bump_size = 1.0;
           float map_Bump_offset[3] = {0, 0, 0};
           float map_Bump_scale[3] = {1, 1, 1};
           )
{
  float amount = 0;
  
  if (map_Bump!="")
  {
    color bc = texture(map_Bump,
                       s*map_Bump_scale[0]+map_Bump_offset[0],
                      (1-t)*map_Bump_scale[1]+map_Bump_offset[1]);
    amount = (comp(bc,0)+comp(bc,1)+comp(bc,2))/3;
  }

  P = P + bump_size*amount*normalize(N);
  N = calculatenormal(P);
}
        """
    
    def displacementShaderParams(self, passes):
        if self.map_Bump==None:
            return {}

        res = {}
        texname = os.path.basename(self.map_Bump.filename)
        name, ext = os.path.splitext(texname)
        res["uniform string map_Bump"] = name+".tex"
        res["uniform float bump_size"] = getattr(self.map_Bump, "bumpsize", 1.0)
        res["uniform float[3] map_Bump_offset"] = self.map_Bump.offset
        res["uniform float[3] map_Bump_scale"] = self.map_Bump.scale
        return res

    def displacementShaderTransform(self):
        return mat4(1)

    def displacementBound(self):
        if self.map_Bump==None:
            return "current", 0
        else:
            return "current", abs(getattr(self.map_Bump, "bumpsize", 1.0))
	
    def interiorShaderName(self):
        return None
    
    def interiorShaderSource(self):
        return None
    
    def interiorShaderParams(self, passes):
        return {}

    def interiorShaderTransform(self):
        return mat4(1)
      
        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: slots.py,v 1.7 2005/08/28 19:55:40 mbaas Exp $

## \file slots.py
## Contains the standard slot classes.

#from _core import ISlot, DoubleSlot, IntSlot, BoolSlot, PySlot
#from _core import Vec3Slot, Vec4Slot, Mat3Slot, Mat4Slot
from .cgtypes import vec3, vec4, mat3, mat4, quat
from . import _core
from ._core import Dependent, UserSizeConstraint, LinearSizeConstraint
from ._core import ISlot, IArraySlot

# Factory functions for the individual slots:

def DoubleSlot(value=0.0, flags=0):
    return _core.DoubleSlot(value, flags)

def IntSlot(value=0, flags=0):
    return _core.IntSlot(value, flags)

def BoolSlot(value=False, flags=0):
    return _core.BoolSlot(value, flags)

def StrSlot(value="", flags=0):
    return _core.StrSlot(value, flags)

def PySlot(value=None, flags=0):
    return _core.PySlot(value, flags)

def Vec3Slot(value=vec3(), flags=0):
    return _core.Vec3Slot(value, flags)

def Vec4Slot(value=vec4(), flags=0):
    return _core.Vec4Slot(value, flags)

def Mat3Slot(value=mat3(), flags=0):
    return _core.Mat3Slot(value, flags)

def Mat4Slot(value=mat4(), flags=0):
    return _core.Mat4Slot(value, flags)

def QuatSlot(value=quat(), flags=0):
    return _core.QuatSlot(value, flags)



def DoubleArraySlot(multiplicity=1, constraint=None):
    return _core.DoubleArraySlot(multiplicity, constraint)

def IntArraySlot(multiplicity=1, constraint=None):
    return _core.IntArraySlot(multiplicity, constraint)

def BoolArraySlot(multiplicity=1, constraint=None):
    return _core.BoolArraySlot(multiplicity, constraint)

def StrArraySlot(multiplicity=1, constraint=None):
    return _core.StrArraySlot(multiplicity, constraint)

def Vec3ArraySlot(multiplicity=1, constraint=None):
    return _core.Vec3ArraySlot(multiplicity, constraint)

def Vec4ArraySlot(multiplicity=1, constraint=None):
    return _core.Vec4ArraySlot(multiplicity, constraint)

#def Mat3ArraySlot(multiplicity=1, constraint=None):
#    return _core.Mat3ArraySlot(multiplicity, constraint)

def Mat4ArraySlot(multiplicity=1, constraint=None):
    return _core.Mat4ArraySlot(multiplicity, constraint)

#def QuatArraySlot(multiplicity=1, constraint=None):
#    return _core.QuatArraySlot(multiplicity, constraint)


# Test

class ProceduralDoubleSlot(_core.DoubleSlot):
    def __init__(self, proc):
        _core.DoubleSlot.__init__(self, 0.0, 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()

class ProceduralIntSlot(_core.IntSlot):
    def __init__(self, proc):
        _core.IntSlot.__init__(self, 0, 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()

class ProceduralVec3Slot(_core.Vec3Slot):
    def __init__(self, proc):
        _core.Vec3Slot.__init__(self, vec3(), 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()

class ProceduralVec4Slot(_core.Vec4Slot):
    def __init__(self, proc):
        _core.Vec4Slot.__init__(self, vec4(), 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()

class ProceduralMat3Slot(_core.Mat3Slot):
    def __init__(self, proc):
        _core.Mat3Slot.__init__(self, mat3(1), 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()

class ProceduralMat4Slot(_core.Mat4Slot):
    def __init__(self, proc):
        _core.Mat4Slot.__init__(self, mat4(1), 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()

class ProceduralQuatSlot(_core.QuatSlot):
    def __init__(self, proc):
        _core.QuatSlot.__init__(self, quat(), 2) # 2 = NO_INPUT_CONNECTIONS
        self._proc = proc

    def computeValue(self):
        self._value = self._proc()


# NotificationForwarder
class NotificationForwarder(_core.Dependent):
    """Forwards slot notifications.

    This class has the same functionality than the C++ class with the
    same name. However, it's not wrapping the C++ class but is pure
    Python.
    """
    def __init__(self, onvalchanged, onresize=None):
        """Constructor.

        If the forwarder is used with an array slot, the onvalchanged
        callback must take two arguments start and end. If it's used with
        a normal slot, then it does not take any arguments.

        \param onvalchanged A callable object that gets called whenever the value changes
        \param onresize A callable object that gets called whenever the size of an array slot changes
        """
        _core.Dependent.__init__(self)
#        self.callable = callable
        self.onvalchanged = onvalchanged
        self.onresize = onresize
        
    def onValueChanged(self, start=None, end=None):
        # Does the call come from a normal slot?
        if start==None:
            self.onvalchanged()
        # or an array slot?
        else:
            self.onvalchanged(start, end)

    def onResize(self, size):
        if self.onresize!=None:
            self.onresize(size)


# slotPropertyCode
def slotPropertyCode(name, slotname=None):
    """Create the code to add a slot property to a class.

    This is a helper function that creates the code string which
    adds a get() method, a set() method and a property to access
    the value of a slot. The resulting string can then be executed
    inside a class definition via an exec statement. The slot itself
    must be present as a variable in the class.

    The returned string looks like this:
    
    \code
    # Property: <name>
    def _get<Name>(self):
        return self.<slotname>.getValue()

    def _set<Name>(self, val):
        self.<slotname>.setValue(val)

    <name> = property(_get<Name>, _set<Name>, None, "<name>")
    \endcode

    In a %component class (or whatever class) the property is created
    as shown in the following example:

    \code
    class Sphere:
        
        exec slotPropertyCode("radius")

        def __init__(self, ...):
            self.radius_slot = ...
    \endcode

    If no slot name is given then the name is assumed to be '<name>_slot'.
    
    \param name (\c str) Name of the new attribute (property)
    \param slotname (\c str) Name of the corresponding slot
    """
    capname = name.capitalize()
    getmethname = "_get%s"%(capname)
    setmethname = "_set%s"%(capname)
    if slotname==None:
        slotname = "%s_slot"%name

    s="""# Property: %s
def %s(self):
    return self.%s.getValue()

def %s(self, val):
    if isinstance(val, _core.Component):
        val.output_slot.connect(self.%s)
    else:
        self.%s.setValue(val)

%s = property(%s, %s, None, "%s")
"""%(name,
     getmethname,slotname,
     setmethname,slotname,slotname,
     name,getmethname,setmethname,capname)
    return s



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: targetcamera.py,v 1.8 2005/07/21 15:07:11 mbaas Exp $

## \file targetcamera.py
## Contains the TargetCamera class.

"""This module contains the TargetCamera class."""

from .Interfaces import *
from . import protocols
from . import slots
from .cgtypes import *
from math import pi
from .worldobject import WorldObject
from .globalscene import getScene
from . import camerabase, lookat
import sys
from . import _core

# TargetCamera
class TargetCamera(camerabase.CameraBase):
    """A camera object that always looks at a particular target.
    """

    protocols.advise(instancesProvide=[ISceneItem, ICamera])

    def __init__(self,
                 name = "TargetCamera",
                 fov = 45.0,
                 target = vec3(0,0,0),
                 roll = 0.0,
                 up = None,
                 fstop = 0,
                 focallength = 0,
                 **params):
        
        camerabase.CameraBase.__init__(self, name=name, **params)

        target = vec3(target)

        # FOV
        self.fov_slot = slots.DoubleSlot(fov)
        self.addSlot("fov", self.fov_slot)

        # Target
        self.target_slot = slots.Vec3Slot(target)
        self.addSlot("target", self.target_slot)

        # Roll
        self.roll_slot = slots.DoubleSlot(roll)
        self.addSlot("roll", self.roll_slot)

        # Up
        self.up_slot = slots.Vec3Slot()
        self.addSlot("up", self.up_slot)
        if up==None:
            self.up_slot.setValue(getScene().up)
        else:
            self.up_slot.setValue(vec3(up))

        self._lookat = lookat.LookAt()
        self._lookat.name = "TargetCamera_LookAt"
        self._lookat.output_slot.connect(self.rot_slot)
        self.pos_slot.connect(self._lookat.pos_slot)
        self.target_slot.connect(self._lookat.target_slot)
        self.roll_slot.connect(self._lookat.roll_slot)
        self.up_slot.connect(self._lookat.up_slot)

        # fstop
        self.fstop_slot = slots.DoubleSlot(fstop)
        self.addSlot("fstop", self.fstop_slot)
        
        # focal length
        self.focallength_slot = slots.DoubleSlot(focallength)
        self.addSlot("focallength", self.focallength_slot)


    def destroy(self):
#        self.fov_slot.setController(None)
#        self.target_slot.setController(None)
#        self._lookat.pos_slot.setController(None)
#        self._lookat.target_slot.setController(None)
#        self.transform_slot.setController(None)
#        self.pos_slot.setController(None)
#        self.rot_slot.setController(None)
#        self.scale_slot.setController(None)
        del self.fov_slot
        del self.target_slot
        del self._lookat.output_slot
        del self._lookat.pos_slot
        del self._lookat.target_slot
        del self.fstop_slot
        del self.focallength_slot

    # projection
    def projection(self, width, height, near, far):
        return mat4().perspective(self.fov, float(width)/height, near, far)

    # viewTransformation
    def viewTransformation(self):
        return self.worldtransform.inverse()      
     
    ## protected:

    exec(slots.slotPropertyCode("fstop"))
    exec(slots.slotPropertyCode("focallength"))
    exec(slots.slotPropertyCode("roll"))
    exec(slots.slotPropertyCode("up"))

    # "fov" property...
    
    def _getFOV(self):
        """Return the current field of view.

        This method is used for retrieving the \a fov property.

        \return Field of view in angles (\c float)
        """
        return self.fov_slot.getValue()

    def _setFOV(self, fov):
        """Set the field of view.

        This method is used for setting the \a fov property.

        \param fov (\c float) Field of view in angles (0-180)
        """
        fov = float(fov)
        if fov<0:
            fov = 0.0
        if fov>180:
            fov = 180.0
        self.fov_slot.setValue(fov)

    fov = property(_getFOV, _setFOV, None, "Field of view (in angles)")
        
    # "target" property...
    
    def _getTarget(self):
        """Return the current target position.

        This method is used for retrieving the \a target property.

        \return Target position (\c vec3)
        """
        return self.target_slot.getValue()

    def _setTarget(self, pos):
        """Set a new target position.

        This method is used for setting the \a target property.

        \param pos (\c vec3) Target position
        """
        pos = vec3(pos)
        self.target_slot.setValue(pos)

    target = property(_getTarget, _setTarget, None, "Target position")



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: spotlight3ds.py,v 1.2 2005/07/03 09:43:36 mbaas Exp $

## \file spotlight3ds.py
## Contains the SpotLight3DS class.

from . import protocols
from .Interfaces import *
from .slots import *
from . import lookat
from .lightsource import LightSource

# SpotLight3DS
class SpotLight3DS(LightSource):
    """This class represents a spotlight as it appears in 3DS files.

    The direction of the light is the local positive Z axis.
    """

    protocols.advise(instancesProvide=[ISceneItem])

    def __init__(self,
                 name="SpotLight3DS",
                 enabled = True,   # off
                 intensity = 1.0,  # multiplier

                 see_cone = False,
                 roll = 0.0, 
                 outer_range = 0,
                 inner_range = 0,
                 attenuation = 0,
                 rectangular_spot = 0,
                 shadowed = False,
                 shadow_bias = 0,
                 shadow_filter = 4.0,
                 shadow_size = 256,
                 spot_aspect = 0,
                 use_projector = False,
                 projector = 0,
                 overshoot = False,
                 ray_shadows = False,
                 ray_bias = False,
                 hotspot = 43,
                 falloff = 45,
                 color = (1,1,1),
                 target=vec3(0,0,0),  # spot
#                 transform = None,
#                 pos = None, rot = None, scale = None,
#                 pivot= None,
#                 offsetTransform = None,
#                 auto_insert=True,
                 **params
                 ):

        LightSource.__init__(self, name=name, **params)

        target = vec3(target)

        # Target
        self.target_slot = Vec3Slot(target)
        self.addSlot("target", self.target_slot)

        self.enabled = enabled
        self.intensity = intensity
        self.attenuation = attenuation
        self.attenuation = attenuation
        self.inner_range = inner_range
        self.outer_range = outer_range
        self.falloff = falloff
        self.hotspot = hotspot
        self.overshoot = overshoot
        self.color = color
        self.shadowed = shadowed
        self.cast_shadow = shadowed
        self.shadow_filter = shadow_filter
        self.shadow_bias = shadow_bias
        self.shadow_size = shadow_size

        # Create the internal LookAt component
        self._lookat = lookat.LookAt()
        self._lookat.name = "SpotLight3DS_LookAt"
        self._lookat.output_slot.connect(self.rot_slot)
        self.pos_slot.connect(self._lookat.pos_slot)
        self.target_slot.connect(self._lookat.target_slot)
        

    # Create the "target" property
    exec(slotPropertyCode("target"))

    def protocols(self):
        return [ISceneItem]

    

        



# -*- coding: utf-8 -*-

__all__ = [
    "__version__",
    "solve",
    "kepler",
]

from .kepler import solve
from .core import kepler
from .kepler_version import __version__

__uri__ = "https://github.com/dfm/kepler.py"
__author__ = "Daniel Foreman-Mackey"
__email__ = "foreman.mackey@gmail.com"
__license__ = "MIT"
__description__ = "Fast and stable Kepler solver"



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: bvhimport.py,v 1.1 2005/02/11 14:32:34 mbaas Exp $

from .cgtypes import *
from .joint import Joint
from .valuetable import ValueTable
from . import bvh
from . import pluginmanager
from .sl import *

# BVHReader
class BVHReader(bvh.BVHReader):
    """Specialized BVH reader class.

    This class creates a hierarchy of joints and applies the motion to it.
    """
    
    def __init__(self, filename):
        bvh.BVHReader.__init__(self, filename)
      

    def onHierarchy(self, root):
        self.createSkeleton(root)
        self.root = root

    def onMotion(self, frames, dt):
        self.frames = frames
        self.dt = dt
        self.currentframe = 0

    def onFrame(self, values):
        self.applyMotion(self.root, values)
        self.currentframe += 1

    def applyMotion(self, node, values):
        """Apply a motion sample to the skeleton.

        node is the current joint and values the joint angles for the
        entire skeleton.
        The method returns the remaining joint angles.
        """

        t = self.currentframe*self.dt
        
        nc = len(node.channels)
        vals = values[:nc]
        pos = vec3()
        pos_flag = False
        for ch,v in zip(node.channels, vals):
            if ch=="Xrotation":
                node.vtx.add(t, v)
            elif ch=="Yrotation":
                node.vty.add(t, v)
            elif ch=="Zrotation":
                node.vtz.add(t, v)
            elif ch=="Xposition":
                pos.x = v
                pos_flag = True
            elif ch=="Yposition":
                pos.y = v
                pos_flag = True
            elif ch=="Zposition":
                pos.z = v
                pos_flag = True

        if pos_flag:
            node.vtpos.add(t, pos)
            
        values = values[nc:]
        for c in node.children:
            values = self.applyMotion(c, values)
        return values

    # createSkeleton
    def createSkeleton(self, node, parent=None):
        """Create the skeleton hierarchy.

        This method creates the skeleton recursively. Each invocation
        creates one joint.
        """
        order = self.rotationOrder(node.channels)
        # Create a new Joint object
        j = Joint(name = node.name,
                  pos = vec3(node.offset),
                  rotationorder = order,
                  parent = parent)
        # Store the joint in the node so that later the motion can be applied
        node.joint = j
        
        vtx = ValueTable(type="double")
        vty = ValueTable(type="double")
        vtz = ValueTable(type="double")
        vtx.output_slot.connect(j.anglex_slot)
        vty.output_slot.connect(j.angley_slot)
        vtz.output_slot.connect(j.anglez_slot)
        node.vtx = vtx
        node.vty = vty
        node.vtz = vtz
        if node.isRoot():
            vtpos = ValueTable(type="vec3")
            vtpos.output_slot.connect(j.pos_slot)
            node.vtpos = vtpos
            
        for c in node.children:
            self.createSkeleton(c, j)

    # rotationOrder
    def rotationOrder(self, channels):
        """Determine rotation order string from the channel names.
        """
        res = ""
        for c in channels:
            if c[-8:]=="rotation":
                res += c[0]

        # Complete the order string if it doesn't already contain
        # all three axes
        m = { "":"XYZ",
              "X":"XYZ", "Y":"YXZ", "Z":"ZXY",
              "XY":"XYZ", "XZ":"XZY",
              "YX":"YXZ", "YZ":"YZX",
              "ZX":"ZXY", "ZY":"ZYX" }
        if res in m:
            res = m[res]
        return res

######################################################################

# BVHImporter
class BVHImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["bvh"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "Biovision Hierarchical"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename):
        """Import a BVH file."""
        
        bvh = BVHReader(filename)
        bvh.read()


######################################################################

# Register the Importer class as a plugin class
pluginmanager.register(BVHImporter)




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2010
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

import sys, os.path, ctypes

def findFfmpegLib(name):
    """Find a ffmpeg library.
    
    name is the name of the library that uses a dot to separate name from
    version number (e.g. "avutil.49").
    The function tries to find the library in some common location and it
    also tries the name where a dot has been replaced by a dash (e.g. "avutil-49").
    
    If the library can't be found, an exception is thrown, otherwise a ctypes
    handle to the library is returned.
    """
    
    # A list of library names that are tried to load the lib
    tries = [name]
    
    if not os.path.isabs(name):
        paths = []
        
        # OSX?
        if sys.platform=="darwin":
            paths = ["/usr/local/lib", "/opt/local/lib", "/sw/lib"]
            for path in paths:
                tries.append(os.path.join(path, "lib%s.dylib"%name))
        # Linux?
        elif sys.platform=="linux2":
            f = name.split(".")
            if len(f)==2:
                name = "lib%s.so.%s"%(f[0], f[1])
                tries.append(name)
            paths = ["/usr/local/lib"]
            for path in paths:
                tries.append(os.path.join(path, name))
        # Windows?
        elif sys.platform=="win32":
            winName = name.replace(".", "-")
            tries.append(winName)
            paths = []

    # Try all the alternatives we have...    
    exc = None
    for libName in tries:
        try:
            lib = ctypes.cdll[libName]
            return lib
        except OSError:
            if exc is None:
                exc = sys.exc_info()[1]
    
    raise exc



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

## \file glfreespotlight.py
## Contains the GLFreeSpotLight class.

from . import protocols
from .Interfaces import *
from .slots import *
from .cgtypes import vec3
from .worldobject import _initWorldObject
from . import cmds
from . import _core

# GLFreeSpotLight
class GLFreeSpotLight(_core.GLSpotLight):
    """This class represents an OpenGL spot light.

    The direction of the light is the local positive Z axis.
    """

    protocols.advise(instancesProvide=[ISceneItem])

    def __init__(self,
                 name="GLSpotLight",
                 parent=None,
                 enabled=True,
                 intensity=1.0,
                 ambient=None,
                 diffuse=None,
                 specular=None,
                 constant_attenuation=1.0,
                 linear_attenuation=0.0,
                 quadratic_attenuation=0.0,
                 exponent=0.0,
                 cutoff=45.0,
                 cast_shadow = False,
                 auto_insert=True,
                 **params
                 ):
        _initWorldObject(self, baseClass=_core.GLSpotLight,
                         name=name, parent=parent,
                         auto_insert=auto_insert,
                         **params)

        self.enabled = enabled
        self.intensity = intensity
        if ambient!=None:
            self.ambient = vec3(ambient)
        if diffuse!=None:
            self.diffuse = vec3(diffuse)
        if specular!=None:
            self.specular = vec3(specular)
        self.constant_attenuation = constant_attenuation
        self.linear_attenuation = linear_attenuation
        self.quadratic_attenuation = quadratic_attenuation
        self.exponent = exponent
        self.cutoff = cutoff
        self.cast_shadow = cast_shadow

    def protocols(self):
        return [ISceneItem]


        



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

from ._OpenGL.GL import *
from . import _Image as Image
from . import _ImageDraw as ImageDraw
import glob, types
from . import component
from .eventmanager import eventManager
from .events import *
from .globalscene import getScene
from .targetcamera import TargetCamera
from .plane import Plane
from .quadrics import Sphere
from .glmaterial import GLMaterial, GLTexture
from .sl import *
from .slots import *
from .cgtypes import *
from math import *
from . import cmds
from . import _core

# Cube transition
class XCube:
    def __init__(self, duration=2.0):
        self.duration = duration
    
    # preTransition
    def preTransition(self, slideshow):
        """Prepare transition.

        This method is called at the beginning of a transition.
        """
        slideshow.backplate.rot = mat3().rotation(pi/2, vec3(0,1,0))
        slideshow.backplate.pos = vec3(2,0,-2)
        slideshow.helper.pos = vec3(0,0,-2)
        slideshow.helper.rot = mat3(1)
        
        cmds.link(slideshow.frontplate, slideshow.helper)
        cmds.link(slideshow.backplate, slideshow.helper)

    # doTransition
    def doTransition(self, slideshow, t):
        """Do the transition.

        This method is called every frame during the transition.
        t is the current transition value (0-1).
        At t=0 the frontplate shows and the backplate is not visible,
        at t=1 the frontplate should not be visible anymore and the
        backplate has come to sight
        """
        w = smoothstep(0,1,smoothstep(0, 1, t))
        slideshow.helper.rot = mat3().rotation(-w*pi/2, vec3(0,1,0))

    # postTransition
    def postTransition(self, slideshow):
        """Post transition.

        This method is called at the end of a transition.
        post: frontplate must be visible
        """
        slideshow.helper.rot = mat3(1)
        cmds.link(slideshow.frontplate, None)
        cmds.link(slideshow.backplate, None)


# Crossfade
class XFade:
    def __init__(self, duration=2.0, zmove=0.0):
        self.duration = duration
        self.zmove = zmove

    ############ XFade transition ####################

    def preTransition(self, slideshow):
        slideshow.backplate.pos = slideshow.frontplate.pos+vec3(0,0,-0.001)
        mat = slideshow.frontplate.getMaterial()
        mat.blend_sfactor = GL_SRC_ALPHA
        mat.blend_dfactor = GL_ONE_MINUS_SRC_ALPHA

    def doTransition(self, slideshow, t):
        w = smoothstep(0,1,t)
        mat = slideshow.frontplate.getMaterial()
        mat.diffuse = vec4(1,1,1,1-w)
        slideshow.frontplate.pos = vec3(0,0,self.zmove*t*t)

    def postTransition(self, slideshow):
        mat = slideshow.frontplate.getMaterial()
        mat.diffuse = vec4(1,1,1,1)
        mat.blend_sfactor = -1
        mat.blend_dfactor = -1
        mat = slideshow.backplate.getMaterial()
        mat.diffuse = vec4(1,1,1,1)

######################################################################

# Slide
class Slide:
    def __init__(self, filepattern, transition=XCube()):
        self.files = glob.glob(filepattern)
        self.transition = transition

######################################################################

# SlideShow
class SlideShow(component.Component):
    """Slide show component.
    """
    
    def __init__(self, name="SlideShow", slides=[], auto_insert=True):
        
        component.Component.__init__(self, name, auto_insert)

        if isinstance(slides, str):
            slides = Slide(slides)

        try:
            len(slides)
        except:
            slides = [slides]

        self.images = []
        for s in slides:
            for f in s.files:
                self.images.append((f, s.transition))

        # Currently displayed image index
        self.imageidx = 0

        self.setup()

        em = eventManager()
        em.connect(STEP_FRAME, self)
        em.connect(LEFT_DOWN, self)
        em.connect(KEY_PRESS, self)

        self.transition = self.images[0][1]

        # Global time when the next transition starts
#        self.transition_start = 3.0
        self.transition_start = 999999.0
        # Slide show state
        self.state = 0


    # setup
    def setup(self):
        """Setup the scene.
        """
        scene = getScene()

        # Set up direction
        scene.up = (0,1,0)

        # Camera
        self.cam = TargetCamera(
            name = "SlideShow_Cam",
            pos = (0,0,5.6),
            fov = 30
        )

        # Set background...
        gradient = Image.new("RGB", (2,128))
        draw = ImageDraw.Draw(gradient)
        colbottom = vec3(0.1, 0.1, 0.3)
        colmid = vec3(0.95,0.95,1)
        coltop = vec3(0.2, 0.2, 0.4)
        colormap = [colbottom, colbottom, colmid, coltop, coltop]
        for y in range(128):
            t = float(y)/128
            c = spline(t, colormap)
            draw.line([0,y, 1,y], fill=(int(c.x*255), int(c.y*255), int(c.z*255)))

        back = Plane(
            lx=4,
            ly=3,
            pos=(0,0,-5),
            scale=2.4,
            material = GLMaterial( diffuse = (0,1,1),
                                   texture = GLTexture(image = gradient,
                                                       mode  = GL_REPLACE))
        )


        # Create plate materials...
        initial = Image.new("RGB", (4,4))
        invY = mat4(1).translate(vec3(0,1,0)).scale(vec3(1,-1,1))
        
        self.mat1 = GLMaterial(
           diffuse = (1,1,1,1),
           texture = GLTexture( image = initial,
                                mode = GL_REPLACE,
#                                size = (512,512),
                                transform = invY,
                                wrap_s = GL_CLAMP,
                                wrap_t = GL_CLAMP
                               )
        )
        self.backmaterial = self.mat1
        self.setBackImage(self.images[0][0])

        self.mat2 = GLMaterial(
           diffuse = (1,1,1,1),
           texture = GLTexture( image = initial,
                                mode = GL_REPLACE,
#                                size = (512,512),
                                transform = invY,
                                wrap_s = GL_CLAMP,
                                wrap_t = GL_CLAMP
                               )
        )
        self.backmaterial = self.mat2
        self.setBackImage(self.images[1][0])

        self.frontmaterial = self.mat1
        self.backmaterial = self.mat2

        # Create plates...
        self.frontplate = Plane(lx = 4, ly = 3, material = self.frontmaterial)

        self.backplate = Plane(lx = 4, ly = 3,
                               pos = (0,0,-0.1),
                               material = self.backmaterial
                               )

        self.helper = Sphere(radius=0.1, pos=(0,0,-1))


    # onLeftDown
    def onLeftDown(self, e):
        if self.state==0:
            self.transition_start = getScene().timer().time

    def onKeyPress(self, e):
        # Page down or Enter?
        if e.keycode==281 or e.keycode==13:
            self.transition_start = getScene().timer().time
        # Page up
        elif e.keycode==280:
            pass
        # q (reset cam)
        elif e.key=="q":
            getScene().up = vec3(0,1,0)
            self.cam.pos = vec3(0,0,5.6)
            self.cam.target = vec3(0,0,0)
            self.cam.fov = 30
            cmds.link(self.cam)
        
    # onStepFrame
    def onStepFrame(self):
        timer = getScene().timer()

        ### State 0: Waiting for the transition
        if self.state==0:
            if timer.time>=self.transition_start:
                self.state = 1

        ### State 1: Begin of transition
        if self.state==1:
            self.transition.preTransition(self)
            eventManager().event("SlidePreTransition", self.images[self.imageidx][0])
            self.state = 2
            
        ### State 2: Transition
        if self.state==2:
            if self.transition.duration>0:
                t = (timer.time-self.transition_start)/self.transition.duration
            else:
                t = 1.0
            if t>1.0:
                t = 1.0
            self.transition.doTransition(self, t)
            eventManager().event("SlideDoTransition", t)
            if t==1.0:
                self.state = 3
                
        ### State 3: Transition is over
        elif self.state==3:
            eventManager().event("SlidePostTransition")
            self.transition.postTransition(self)
            self.switchSlide()

            self.frontplate.transform = mat4(1)
            self.backplate.transform = mat4(1)
            self.backplate.pos = vec3(0,0,-1)
                       
#            self.transition_start += 5.0
            self.transition_start = 999999.0
            self.state = 0

    # switchSlide
    def switchSlide(self):
        """Prepare everything for the next slide.
        """
        # Swap front and back material...
        dummy = self.frontmaterial
        self.frontmaterial = self.backmaterial
        self.backmaterial = dummy

        # Update the materials of the plates
        self.frontplate.setMaterial(self.frontmaterial)
        self.backplate.setMaterial(self.backmaterial)

        # Set the next image on the back material
        idx = (self.imageidx+2)%len(self.images)
        self.setBackImage(self.images[idx][0])
        self.imageidx = (self.imageidx+1)%len(self.images)

        # Set the new transition object
        self.transition = self.images[self.imageidx][1]
        

    # setBackImage
    def setBackImage(self, name):
        """Set a new image on the back material."""
        self.backmaterial.texture.imagename = name
        img = Image.open(name)
        if img.mode!="RGB" and img.mode!="RGBA":
            img = img.convert("RGB")
        self.backmaterial.texture.image = img
        w,h = img.size
        f = (4.0*h)/(3.0*w)
        if f>=1.0:
            S = mat4(1).scale(vec3(f,1,1))
            S.translate(vec3(-0.5+0.5/f,0,0))
            draw = ImageDraw.Draw(img)
            draw.line([(0,0), (0,h)], fill=(0,0,0))
            draw.line([(1,0), (1,h)], fill=(0,0,0))
            draw.line([(w-1,0), (w-1,h)], fill=(0,0,0))
            draw.line([(w-2,0), (w-2,h)], fill=(0,0,0))
        else:
            f = (3.0*w)/(4.0*h)
            S = mat4(1).scale(vec3(1,f,1))
            S.translate(vec3(0,-0.5+0.5/f,0))
            draw = ImageDraw.Draw(img)
            draw.line([(0,0), (w,0)], fill=(0,0,0))
            draw.line([(0,1), (w,1)], fill=(0,0,0))
            draw.line([(0,h-1), (w,h-1)], fill=(0,0,0))
            draw.line([(0,h-2), (w,h-2)], fill=(0,0,0))
        invY = mat4(1).translate(vec3(0,1,0)).scale(vec3(1,-1,1))
        self.backmaterial.texture.transform = invY*S
               
            

######################################################################

#sl = SlideShow()



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: beziercurvegeom.py,v 1.2 2005/09/02 12:58:02 mbaas Exp $

from ._OpenGL.GL import *
import bisect
from .geomobject import *
from .slots import *
from .cgtypes import *
from .boundingbox import BoundingBox
from . import protocols
from .ribexport import IGeometry
from .ri import *

# BezierPoint
class BezierPoint:
    """A single Bezier point with tangents.

    This object is used during construction of a BezierCurveGeom object.
    """
    def __init__(self, pos, intangent=vec3(0), outtangent=vec3(0)):
        self.pos = vec3(pos)
        self.intangent = vec3(intangent)
        self.outtangent = vec3(outtangent)

_uniform_size_constraint = UserSizeConstraint(1)
_null_size_constraint = UserSizeConstraint(0)

# BezierCurveGeom
class BezierCurveGeom(GeomObject):
    """Cubic Bezier curve.
    """

    protocols.advise(instancesProvide=[IGeometry])
    
    def __init__(self,
                 pnts = None,
                 closed = False,
                 epsilon = 0.01,
                 subdiv = 4,
                 show_tangents = False):
        """Constructor.

        pnts is a list of BezierPoint objects.
        """
        GeomObject.__init__(self)

        # Epsilon value for segLength()
        self.epsilon = epsilon
        # Number of subdivisions for drawGL()
        self.subdiv = subdiv
        # Show tangents or not?
        self.show_tangents = show_tangents

        # Create the slots...
        
        # The end points of the Bezier segments, i.e. the curve passes
        # through these points.
        self.pnts_slot = Vec3ArraySlot()
        self._var_sizeconstraint = LinearSizeConstraint(self.pnts_slot,1,0)
        if closed:
            b = 0
        else:
            b = -2
        self._vtx_sizeconstraint = LinearSizeConstraint(self.pnts_slot,3,b)
        # The incoming tangent for point i
        self.intangents_slot = Vec3ArraySlot(constraint=self._var_sizeconstraint)
        # The outgoing tangent for point i
        self.outtangents_slot = Vec3ArraySlot(constraint=self._var_sizeconstraint)
        
        # Create aliases for the slot "value" attribute
        self.pnts = self.pnts_slot
        self.intangents = self.intangents_slot
        self.outtangents = self.outtangents_slot

        # Add the slots to the component
        self.addSlot("pnts", self.pnts_slot)
        self.addSlot("intangents", self.intangents_slot)
        self.addSlot("outtangents", self.outtangents_slot)

        self._closed = closed

        if pnts!=None:
            # Initialize the Bezier points...
            self.pnts.resize(len(pnts))
            for i,bp in enumerate(pnts):
                self.pnts[i] = bp.pos
                self.intangents[i] = bp.intangent
                self.outtangents[i] = bp.outtangent

        # A list of lengths (the lengths are accumulated, so it's
        # the length of the entire curve until segment i)
        self._seg_lengths = []
        # Flag that indicates if all internal precomputed attributes have
        # to be recomputed or if they are still valid.
        self._internal_data_invalid = True

        # Have the slots call the onCurveChanged() method whenever
        # a point or a tangent is modified.
        self._nf = NotificationForwarder(self.onCurveChanged, self.onCurveResized)
        self.pnts_slot.addDependent(self._nf)
        self.intangents_slot.addDependent(self._nf)
        self.outtangents_slot.addDependent(self._nf)


    def _updateInternalData(self):
        """Precalculate the lengths of the Bezier segments.
        """
        self._seg_lengths = []
        sl = 0.0
        if self._closed:
            k = 0
        else:
            k = 1
        for i in range(self.pnts.size()-k):
            sl += self.segLength(self.segCtrlPoints(i), self.epsilon)
            self._seg_lengths.append(sl)

        self._internal_data_invalid = False
        

#    def uniformCount(self):
#        return 1

#    def varyingCount(self):
#        # One value per segment boundary, i.e.
#        # self.numsegs   for periodic curves and
#        # self.numsegs+1 for non-periodic curves
#        # which can be reduced to self.pnts.size() for both cases
#        # (as the curve points just mark segment boundaries)
#        return self.pnts.size()

#    def vertexCount(self):
#        # One value per control points,
#        n = 3*self.numsegs
#        if not self._closed:
#            n += 1
#        return n

    # slotSizeConstraint
    def slotSizeConstraint(self, storage):
        global _uniform_size_constraint
        
        if storage==UNIFORM:
            return _uniform_size_constraint
        elif storage==VARYING:
            return self._var_sizeconstraint
        elif storage==VERTEX:
            return self._vtx_sizeconstraint
        else:
            return _null_size_constraint

    # boundingBox
    def boundingBox(self):
        """Return the bounding box of the control polygon.
        """
        bb = BoundingBox()
        for i,p in enumerate(self.pnts):
            bb.addPoint(p)
            bb.addPoint(p+self.intangents[i])
            bb.addPoint(p+self.outtangents[i])
        return bb

    # drawGL
    def drawGL(self):
        if self.pnts.size()==0:
            return

        glPushAttrib(GL_LIGHTING_BIT)
        glDisable(GL_LIGHTING)
        glColor3f(1,1,1)

        # Draw the curve...
        p0 = self.pnts[0]
        in0 = self.intangents[0]
        out0 = self.outtangents[0]
        for i in range(1, self.pnts.size()):
            p1 = self.pnts[i]
            in1 = self.intangents[i]
            out1 = self.outtangents[i]
            self.segDrawGL([p0, p0+out0, p1+in1, p1], self.subdiv)
            p0 = p1
            in0 = in1
            out0 = out1

        # Draw an additional segment if the curve is closed...
        if self._closed:
            p0 = self.pnts[-1]
            out0 = self.outtangents[-1]
            p1 = self.pnts[0]
            in1 = self.intangents[0]
            self.segDrawGL([p0, p0+out0, p1+in1, p1], self.subdiv)
            

        # Draw the in and out tangents...
        if self.show_tangents:
            for i in range(self.pnts.size()):
                p = self.pnts[i]
                pin = self.intangents[i]
                pout = self.outtangents[i]
                glBegin(GL_LINE_STRIP)
                glColor3f(0,1,0)
                v = p+pin
                glVertex3d(v.x, v.y, v.z)
                glVertex3d(p.x, p.y, p.z)
                glColor3f(0,0,1)
                glVertex3d(p.x, p.y, p.z)
                v = p+pout
                glVertex3d(v.x, v.y, v.z)
                glEnd()            

        glPopAttrib()

    # render
    def render(self, matid):
        if matid==0:
            # Set Bezier basis
            RiBasis(RiBezierBasis, RI_BEZIERSTEP, RiBezierBasis, RI_BEZIERSTEP)

            # Create the list of curve vertices...
            pnts = []
            for i in range(self.numsegs):
                ps = self.segCtrlPoints(i)
                pnts += ps[0:3]

            if self._closed:
                wrap = RI_PERIODIC
            else:
                wrap = RI_NONPERIODIC
                pnts.append(self.pnts[-1])

            # Create parameter list...
            params = {"P":pnts}
            clss = ["constant", "uniform", "varying", "vertex",
                    "facevarying", "facevertex", "user"]
            typs = ["integer", "float", "string", "color", "point",
                    "vector", "normal", "matrix", "hpoint"]
            for name, storage, type, multiplicity in self.iterVariables():
                cls = clss[abs(storage)]
                typ = typs[abs(type)]
                if cls=="user":
                    continue
                s = self.slot(name)
                params[name] = list(s)
                if multiplicity==1:
                    decl = "%s %s"%(cls, typ)
                else:
                    decl = "%s %s[%d]"%(cls, typ, multiplicity)
                RiDeclare(name, decl)

            # Output a curve primitive
            RiCurves(RI_CUBIC, [len(pnts)], wrap, params)

    # eval
    def eval(self, t):
        """Evaluate the curve at parameter t and return the curve point.
        """
        # Segment number
        seg = int(t)
        # Segment parameter
        t = t%1.0
#        if seg>=self.pnts.size()-1:
#            seg = self.pnts.size()-2
#            t = 1.0
        return self.segEval(t, self.segCtrlPoints(seg))        

    # evalFrame
    def evalFrame(self, t):
        """Evaluate the curve at parameter t and return a coordinate frame.
        """
        # Segment number
        seg = int(t)
        # Segment parameter
        t = t%1.0
#        if seg>=self.pnts.size()-1:
#            seg = self.pnts.size()-2
#            t = 1.0
        return self.segEvalFrame(t, self.segCtrlPoints(seg))

    # deriv
    def deriv(self, t):
        """Return the derivative at parameter t.
        """
        # Segment number
        seg = int(t)
        # Segment parameter
        t = t%1.0
#        if seg>=self.pnts.size()-1:
#            seg = self.pnts.size()-2
#            t = 1.0
        return self.segDeriv(t, self.segCtrlPoints(seg))

    # arcLen
    def arcLen(self, t):
        """Return the arc length up to t.
        """
        if self._internal_data_invalid:
            self._updateInternalData()
            
        seg = int(t)
        t = t%1.0
        if seg>=len(self._seg_lengths):
            return self._seg_lengths[-1]

        res = 0.0
        if seg>0:
            res = self._seg_lengths[seg-1]
        b,c = self.segSplit(self.segCtrlPoints(seg), t)
        return res + self.segLength(b, self.epsilon)

    # length
    def length(self):
        """Return the length of the entire curve.
        """
        if self._internal_data_invalid:
            self._updateInternalData()
        if len(self._seg_lengths)==0:
            return 0.0
        else:
            return self._seg_lengths[-1]


    def eval0(self, t):
        if self._internal_data_invalid:
            self._updateInternalData()

        # Determine the Bezier segment index we're currently in
        seg = bisect.bisect_right(self._seg_lengths, t)
        if seg==self.pnts.size()-1:
            return self.pnts[-1], vec3(), vec3()

        if seg>0:
            t -= self._seg_lengths[seg-1]

        p,t = self.segEval0(t, self.segCtrlPoints(seg))
        if p!=None:
            return p, vec3(), vec3()
        else:
            return self.pnts[-1], vec3(), vec3()

    # onCurveChanged
    def onCurveChanged(self, start, end):
        """This method is called whenever a point or tangent is modified.
        """
        self._internal_data_invalid = True

    # onCurveResized
    def onCurveResized(self, size):
        """This method is called whenever the number of points is modified.
        """
        self._internal_data_invalid = True
        
    ## protected:

    def segSmooth(self, seg):
        b0,b1,b2,b3 = self.segCtrlPoints(seg-1)
        d1 = b2-b1
        d2 = b3-b2
        self.outtangents[seg] = d2
        self.intangents[seg+1] = self.pnts[seg]+3*d2-d1 - self.pnts[seg+1]

    def segCtrlPoints(self, i):
        """Return the Bezier control points of segment i.

        i may be an arbitrary segment number (the numbering is cyclic).
        """
        s = self.pnts.size()
        i = i%s
        j = (i+1)%s
        b0 = self.pnts[i]
        b3 = self.pnts[j]
        b1 = b0+self.outtangents[i]
        b2 = b3+self.intangents[j]
        return [b0, b1, b2, b3]

    # segEval0
    def segEval0(self, t, ctrlpnts, eps=0.001):
        """

        Return value: (pnt, t2)
        pnt is None if t>length. t2 is t-length.
        """
        l1 = self.segHullLength(ctrlpnts)
        b,c = self.segSplit(ctrlpnts)
        bl = self.segHullLength(b)
        cl = self.segHullLength(c)
        l2 = bl+cl
        # Is the result accurate enough? then return the current result,
        # otherwise subdivide further
        if l1-l2<=eps:
            p,t = self.segHullEval0(t, b)
            if p!=None:
                return p,t
            p,t = self.segHullEval0(t, c)
            if p!=None:
                return p,t
            return None, t
        else:
            p,t = self.segEval0(t, b, eps=eps)
            if p!=None:
                return p,t
            p,t = self.segEval0(t, c, eps=eps)
            if p!=None:
                return p,t
            return None, t

    def segHullEval0(self, t, ctrlpnts):
        b0,b1,b2,b3 = ctrlpnts
        # 1st segment
        d = (b1-b0).length()
        if t<d:
            a = t/d
            return (1-a)*b0 + a*b1, t
        t-=d
        # 2nd segment
        d = (b2-b1).length()
        if t<d:
            a = t/d
            return (1-a)*b1 + a*b2, t
        t-=d
        # 3rd segment
        d = (b3-b2).length()
        if t<d:
            a = t/d
            return (1-a)*b2 + a*b3, t
        t-=d
        return None, t
        
    # segEval
    def segEval(self, t, ctrlpnts):
        """Evaluate the curve at parameter t using the de Casteljau scheme.

        t ranges from 0 to 1.
        """
        _t = 1.0-t
        b0,b1,b2,b3 = ctrlpnts
        c0 = _t*b0 + t*b1
        c1 = _t*b1 + t*b2
        c2 = _t*b2 + t*b3
        d0 = _t*c0 + t*c1
        d1 = _t*c1 + t*c2
        return _t*d0+t*d1

    # segEvalFrame
    def segEvalFrame(self, t, ctrlpnts):
        """Evaluate the curve at parameter t using the de Casteljau scheme.

        t ranges from 0 to 1.
        """
        _t = 1.0-t
        b0,b1,b2,b3 = ctrlpnts
        c0 = _t*b0 + t*b1
        c1 = _t*b1 + t*b2
        c2 = _t*b2 + t*b3
        d0 = _t*c0 + t*c1
        d1 = _t*c1 + t*c2
        e0 = _t*d0+t*d1
        return e0, 3*(d1-d0), 6*(c2-2*c1+c0)

    # segDeriv
    def segDeriv(self, t, ctrlpnts):
        """Return the derivation at t.

        t ranges from 0 to 1.
        """
        _t = 1.0-t
        b0,b1,b2,b3 = ctrlpnts
        c0 = _t*b0 + t*b1
        c1 = _t*b1 + t*b2
        c2 = _t*b2 + t*b3
        d0 = _t*c0 + t*c1
        d1 = _t*c1 + t*c2
        return 3*(d1-d0)


    # segDrawGL
    def segDrawGL(self, ctrlpnts, subdiv=4):
        """Draw one Bezier segment.

        Actually the control polygon is drawn after a number of
        subdivisions.
        """
        if subdiv==0:
            b0,b1,b2,b3 = ctrlpnts
            glBegin(GL_LINE_STRIP)
            glVertex3d(b0.x, b0.y, b0.z)
            glVertex3d(b1.x, b1.y, b1.z)
            glVertex3d(b2.x, b2.y, b2.z)
            glVertex3d(b3.x, b3.y, b3.z)
            glEnd()
        else:
            b,c = self.segSplit(ctrlpnts)
            self.segDrawGL(b, subdiv=subdiv-1)
            self.segDrawGL(c, subdiv=subdiv-1)
            
    # segLength
    def segLength(self, ctrlpnts, eps=0.001):
        """Return the length of one Bezier segment.

        ctrlpnts is a list of four vec3s.
        """
        l1 = self.segHullLength(ctrlpnts)
        b,c = self.segSplit(ctrlpnts)
        bl = self.segHullLength(b)
        cl = self.segHullLength(c)
        l2 = bl+cl
        # Is the result accurate enough? then return the current result,
        # otherwise subdivide further
        if l1-l2<=eps:
            return l2
        else:
            return self.segLength(b,eps)+self.segLength(c,eps)

    # segHullLength
    def segHullLength(self, ctrlpnts):
        """Return the length of the control "polygon".
        """
        b0,b1,b2,b3 = ctrlpnts
        return (b1-b0).length() + (b2-b1).length() + (b3-b2).length()

    # segSplit
    def segSplit(self, ctrlpnts, t=0.5):
        """Split a segment into two segments.
        """
        b0,b1,b2,b3 = ctrlpnts
        _t = 1.0-t
        c0 = _t*b0 + t*b1
        c1 = _t*b1 + t*b2
        c2 = _t*b2 + t*b3
        d0 = _t*c0 + t*c1
        d1 = _t*c1 + t*c2
        e0 = _t*d0 + t*d1
        return [b0,c0,d0,e0], [e0,d1,c2,b3]

    # "paraminterval" property...
    
    def _getParamInterval(self):
        """Return the current parameter interval.

        This method is used for retrieving the \a paraminterval property.

        \return (t_min, t_max)
        """
        if self._closed:
            return 0, self.pnts.size()
        else:
            return 0, self.pnts.size()-1

    paraminterval = property(_getParamInterval, None, None, "Parameter range")

    # "closed" property...
    
    def _getClosed(self):
        return self._closed

    def _setClosed(self, c):
        if c==self._closed:
            return
        self._closed = c
        self._internal_data_invalid = True
        # Update the size constraint for vertex variables
        if self._closed:
            self._vtx_sizeconstraint.setCoeffs(3,0)
        else:
            self._vtx_sizeconstraint.setCoeffs(3,-2)

    closed = property(_getClosed, _setClosed, None, "Closed")

    # "numsegs" property...
    
    def _getNumSegs(self):
        if self._closed:
            return self.pnts.size()
        else:
            return self.pnts.size()-1

    numsegs = property(_getNumSegs, None, None, "Number of Bezier segments")





# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: preferences.py,v 1.1.1.1 2004/12/12 14:31:13 mbaas Exp $

## \file preferences.py
## \brief Contains the Preferences class.

import sys, os, os.path, pickle

# Preferences
class Preferences(object):
    """This class stores arbitrary values persistently.

    This class can be used just like a dictionary.
    """

    def __init__(self, filename):
        """Constructor."""

        # If filename is no absolute path make it relative to the default
        # config file path
        if not os.path.isabs(filename):
            filename = os.path.join(configPath(), filename)
            
        self._prefs = {}
        self._filename = filename

    def __len__(self):
        return len(self._prefs)

    def __iter__(self):
        return iter(list(self._prefs.items()))

    def __getitem__(self, key):
        if key in self._prefs:
            return self._prefs[key]
        else:
            return None

    def __setitem__(self, key, value):
        self._prefs[key]=value

    def __delitem__(self, key):
        if key in self._prefs:
            del self._prefs[key]

    def __contains__(self, item):
        return item in self._prefs

    def get(self, key, default=None):
        return self._prefs.get(key, default)

    # load
    def load(self):
        """Load the preferences.

        """

        f = open(self._filename)
        id = pickle.load(f)
        if id!=1:
            raise Exception("Unknown config file format")
        self._prefs = pickle.load(f)
        f.close()


    # save
    def save(self):
        """Save the preferences.

        """
        
        self._preparePath(os.path.dirname(self._filename))
        
        f = open(self._filename, "w")
        pickle.dump(1, f)
        pickle.dump(self._prefs, f)
        f.close()


    ######################################################################
    ## protected:

    def _preparePath(self, path):
        """Prepare a path so that every directory on the path exists.

        Checks if the path exists and creates it if it does not exist.
        """
        if not os.path.exists(path):
            parent = os.path.dirname(path)
            if parent!="":
                self._preparePath(parent)
            os.mkdir(path)        

    # "filename" property

    def _getFilename(self):
        """Return the filename.

        This method is used for retrieving the \a filename property.

        \return Filename (\c str).
        """
        return self._filename

    filename = property(_getFilename, None, None, "File name")
    

######################################################################

def configPath(appname="gaia"):
    """Return the full path where config files are located.

    \todo Change "gaia" as default application name
    """
    
    # Windows? (XP)
    if sys.platform=="win32":
        if "APPDATA" in os.environ:
            return os.path.abspath(os.path.join(os.environ["APPDATA"], appname.capitalize()))
        else:
            return None
    # Other
    else:
        if "HOME" in os.environ:
            return os.path.abspath(os.environ["HOME"])
        else:
            return None


######################################################################

if __name__=="__main__":

    print((configPath()))
            
            

    



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: vec4.py,v 1.1 2005/08/15 15:39:48 mbaas Exp $

import math

# Comparison threshold
_epsilon = 1E-12


# vec4
class vec4(object):
    """Four-dimensional vector.

    This class represents a 4D vector.
    """

    def __init__(self, *args):
        """Constructor.

        There are several possibilities how to initialize a vector:

        v = vec4()        -> v = <0,0,0,0>
        v = vec4(a)       -> v = <a,a,a,a>
        v = vec4(x,y)     -> v = <x,y,0,0>
        v = vec4(x,y,z)   -> v = <x,y,z,0>
        v = vec4(x,y,z,w) -> v = <x,y,z,w>

        Note that specifying just one value sets all four components to
        that value.

        Additionally you can wrap those values in a list or a tuple or
        specify them as a string:

        v = vec4([1,2,3]) -> v = <1,2,3,0>
        v = vec4("4,5")   -> v = <4,5,0,0>        
        """
        
        if len(args)==0:
            self.x, self.y, self.z, self.w = (0.0, 0.0, 0.0, 0.0)

        elif len(args)==1:
            T = type(args[0])
            # scalar
            if T in [float, int]:
                f = float(args[0])
                self.x, self.y, self.z, self.w = (f, f, f, f)
            # vec4
            elif isinstance(args[0], vec4):
                self.x, self.y, self.z, self.w = args[0]
            # Tuple/List
            elif T in [tuple, list]:
                if len(args[0])==0:
                    self.x = self.y = self.z = self.w = 0.0
                elif len(args[0])==1:
                    self.x = self.y = self.z = self.w = float(args[0][0])
                elif len(args[0])==2:
                    self.x = float(args[0][0])
                    self.y = float(args[0][1])
                    self.z = 0.0
                    self.w  = 0.0
                elif len(args[0])==3:
                    x,y,z = args[0]
                    self.x = float(x)
                    self.y = float(y)
                    self.z = float(z)
                    self.w = 0.0
                elif len(args[0])==4:
                    x,y,z,w = args[0]
                    self.x = float(x)
                    self.y = float(y)
                    self.z = float(z)
                    self.w = float(w)
                else:
                    raise TypeError("vec4() takes at most 4 arguments")
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f=list([float(x) for x in s])
                dummy = vec4(f)
                self.x, self.y, self.z, self.w = dummy
            # error
            else:
                raise TypeError("vec4() arg can't be converted to vec4")

        elif len(args)==2:
            self.x, self.y = (float(args[0]), float(args[1]))
            self.z, self.w = (0.0, 0.0)
            
        elif len(args)==3:
            x,y,z = args
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)
            self.w = 0.0

        elif len(args)==4:
            x,y,z,w = args
            self.x = float(x)
            self.y = float(y)
            self.z = float(z)
            self.w = float(w)

        else:
            raise TypeError("vec4() takes at most 4 arguments")


    def __repr__(self):
        return 'vec4(%r, %r, %r, %r)'%(self.x, self.y, self.z, self.w)

    def __str__(self):
        return '(%1.4f, %1.4f, %1.4f, %1.4f)'%(self.x, self.y, self.z, self.w)

    # Make the object unhashable (as it is mutable)
    __hash__ = None

    def __eq__(self, other):
        """== operator

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.6)
        >>> c=vec4(-0.3, 0.75, 0.5, 0.6)
        >>> print a==b
        0
        >>> print b==c
        1
        >>> print a==None
        0
        """
        global _epsilon
        if isinstance(other, vec4):
            return (abs(self.x-other.x)<=_epsilon and
                    abs(self.y-other.y)<=_epsilon and
                    abs(self.z-other.z)<=_epsilon and
                    abs(self.w-other.w)<=_epsilon)
        else:
            return False

    def __ne__(self, other):
        """!= operator

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.6)
        >>> c=vec4(-0.3, 0.75, 0.5, 0.6)
        >>> print a!=b
        1
        >>> print b!=c
        0
        >>> print a!=None
        1
        """
        return not (self==other)

    def __lt__(self, other):
        """< operator."""
        if isinstance(other, vec4):
            return ((self.x<other.x) and
                    (self.y<other.y) and
                    (self.z<other.z) and
                    (self.w<other.w))
        else:
            return False

    def __le__(self, other):
        """<= operator."""
        if isinstance(other, vec4):
            return ((self.x<=other.x) and
                    (self.y<=other.y) and
                    (self.z<=other.z) and
                    (self.w<=other.w))
        else:
            return False
        
    def __gt__(self, other):
        """> operator."""
        if isinstance(other, vec4):
            return ((self.x>other.x) and
                    (self.y>other.y) and
                    (self.z>other.z) and
                    (self.w>other.w))
        else:
            return False
        
    def __ge__(self, other):
        """>= operator."""
        if isinstance(other, vec4):
            return ((self.x>=other.x) and
                    (self.y>=other.y) and
                    (self.z>=other.z) and
                    (self.w>=other.w))
        else:
            return False


    def __add__(self, other):
        """Vector addition.

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.3)
        >>> print a+b
        (0.7000, 1.2500, -1.3000, 0.5000)
        """
        if isinstance(other, vec4):
            return vec4(self.x+other.x, self.y+other.y, self.z+other.z, self.w+other.w)
        else:
            raise TypeError("unsupported operand type for +")

    def __sub__(self, other):
        """Vector subtraction.

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.3)
        >>> print a-b
        (1.3000, -0.2500, -2.3000, -0.1000)
        """
        if isinstance(other, vec4):
            return vec4(self.x-other.x, self.y-other.y, self.z-other.z, self.w-other.w)
        else:
            raise TypeError("unsupported operand type for -")

    def __mul__(self, other):
        """Multiplication with a scalar or dot product.

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.3)
        >>> print a*2.0
        (2.0000, 1.0000, -3.6000, 0.4000)
        >>> print 2.0*a
        (2.0000, 1.0000, -3.6000, 0.4000)
        >>> print a*b
        -0.765
        """

        T = type(other)
        # vec4*scalar
        if T in [float, int]:
            return vec4(self.x*other, self.y*other, self.z*other, self.w*other)
        # vec4*vec4
        if isinstance(other, vec4):
            return self.x*other.x + self.y*other.y + self.z*other.z + self.w*other.w
        # unsupported
        else:
            # Try to delegate the operation to the other operand
            if getattr(other,"__rmul__",None)!=None:
                return other.__rmul__(self)
            else:
                raise TypeError("unsupported operand type for *")

    __rmul__ = __mul__

    def __truediv__(self, other):
        """Division by scalar

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> print a/2.0
        (0.5000, 0.2500, -0.9000, 0.1000)
        """
        T = type(other)
        # vec4/scalar
        if T in [float, int]:
            return vec4(self.x/other, self.y/other, self.z/other, self.w/other)
        # unsupported
        else:
            raise TypeError("unsupported operand type for /")

    def __mod__(self, other):
        """Modulo (component wise)

        >>> a=vec4(3.0, 2.5, -1.8, 0.2)
        >>> print a%2.0
        (1.0000, 0.5000, 0.2000, 0.2000)
        """
        T = type(other)
        # vec4%scalar
        if T in [float, int]:
            return vec4(self.x%other, self.y%other, self.z%other, self.w%other)
        # vec4%vec4
        if isinstance(other, vec4):
            return vec4(self.x%other.x, self.y%other.y, self.z%other.z, self.w%other.w)
        # unsupported
        else:
            raise TypeError("unsupported operand type for %")

    def __iadd__(self, other):
        """Inline vector addition.

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.3)
        >>> a+=b
        >>> print a
        (0.7000, 1.2500, -1.3000, 0.5000)
        """
        if isinstance(other, vec4):
            self.x+=other.x
            self.y+=other.y
            self.z+=other.z
            self.w+=other.w
            return self
        else:
            raise TypeError("unsupported operand type for +=")

    def __isub__(self, other):
        """Inline vector subtraction.

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> b=vec4(-0.3, 0.75, 0.5, 0.3)
        >>> a-=b
        >>> print a
        (1.3000, -0.2500, -2.3000, -0.1000)
        """
        if isinstance(other, vec4):
            self.x-=other.x
            self.y-=other.y
            self.z-=other.z
            self.w-=other.w
            return self
        else:
            raise TypeError("unsupported operand type for -=")

    def __imul__(self, other):
        """Inline multiplication (only with scalar)

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> a*=2.0
        >>> print a
        (2.0000, 1.0000, -3.6000, 0.4000)
        """
        T = type(other)
        # vec4*=scalar
        if T in [float, int]:
            self.x*=other
            self.y*=other
            self.z*=other
            self.w*=other
            return self
        else:
            raise TypeError("unsupported operand type for *=")

    def __itruediv__(self, other):
        """Inline division with scalar

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> a/=2.0
        >>> print a
        (0.5000, 0.2500, -0.9000, 0.1000)
        """
        T = type(other)
        # vec4/=scalar
        if T in [float, int]:
            self.x/=other
            self.y/=other
            self.z/=other
            self.w/=other
            return self
        else:
            raise TypeError("unsupported operand type for /=")

    def __imod__(self, other):
        """Inline modulo

        >>> a=vec4(3.0, 2.5, -1.8, 0.2)
        >>> a%=2.0
        >>> print a
        (1.0000, 0.5000, 0.2000, 0.2000)
        """
        T = type(other)
        # vec4%=scalar
        if T in [float, int]:
            self.x%=other
            self.y%=other
            self.z%=other
            self.w%=other
            return self
        # vec4%vec4
        if isinstance(other, vec4):
            self.x%=other.x
            self.y%=other.y
            self.z%=other.z
            self.w%=other.w
            return self
        else:
            raise TypeError("unsupported operand type for %=")

    def __neg__(self):
        """Negation

        >>> a=vec4(3.0, 2.5, -1.8, 0.2)
        >>> print -a
        (-3.0000, -2.5000, 1.8000, -0.2000)
        """
        return vec4(-self.x, -self.y, -self.z, -self.w)

    def __pos__(self):
        """
        >>> a=vec4(3.0, 2.5, -1.8, 0.2)
        >>> print +a
        (3.0000, 2.5000, -1.8000, 0.2000)
        """
        return vec4(+self.x, +self.y, +self.z, +self.w)

    def __abs__(self):
        """Return the length of the vector.

        abs(v) is equivalent to v.length().

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> print abs(a)
        2.12837966538
        """
        return math.sqrt(self*self)


    def __len__(self):
        """Length of the sequence (always 4)"""
        return 4

    def __getitem__(self, key):
        """Return a component by index (0-based)

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> print a[0]
        1.0
        >>> print a[1]
        0.5
        >>> print a[2]
        -1.8
        >>> print a[3]
        0.2
        """
        T=type(key)
        if T is not int:
            raise TypeError("index must be integer")

        if   key==0: return self.x
        elif key==1: return self.y
        elif key==2: return self.z
        elif key==3: return self.w
        else:
            raise IndexError("index out of range")

    def __setitem__(self, key, value):
        """Set a component by index (0-based)

        >>> a=vec4()
        >>> a[0]=1.5; a[1]=0.7; a[2]=-0.3; a[3]=0.2
        >>> print a
        (1.5000, 0.7000, -0.3000, 0.2000)
        """
        T=type(key)
        if T is not int:
            raise TypeError("index must be integer")

        if   key==0: self.x = value
        elif key==1: self.y = value
        elif key==2: self.z = value
        elif key==3: self.w = value
        else:
            raise IndexError("index out of range")


    def length(self):
        """Return the length of the vector.

        v.length() is equivalent to abs(v).

        >>> a=vec4(1.0, 0.5, -1.8, 0.2)
        >>> print a.length()
        2.12837966538
        """

        return math.sqrt(self*self)

    def normalize(self):
        """Return normalized vector.

        >>> a=vec4(1.0, 0.5, -1.8, 1.2)
        >>> print a.normalize()
        (0.4107, 0.2053, -0.7392, 0.4928)
        """

        nlen = 1.0/math.sqrt(self*self)
        return vec4(self.x*nlen, self.y*nlen, self.z*nlen, self.w*nlen)

    def min(self):
        """Return the minimum value of the components.
        """
        return min(self.x, self.y, self.z, self.w)

    def max(self):
        """Return the maximum value of the components.
        """
        return max(self.x, self.y, self.z, self.w)

    def minIndex(self):
        """Return the index of the component with the minimum value.
        """
        if self.x<=self.y and self.x<=self.z and self.x<=self.w:
            return 0
        elif self.y<=self.z and self.y<=self.w:
            return 1
        elif self.z<=self.w:
            return 2
        else:
            return 3

    def maxIndex(self):
        """Return the index of the component with the maximum value.
        """
        if self.x>=self.y and self.x>=self.z and self.x>=self.w:
            return 0
        elif self.y>=self.z and self.y>=self.w:
            return 1
        elif self.z>=self.w:
            return 2
        else:
            return 3
        
    def minAbs(self):
        """Return the minimum absolute value of the components.
        """
        return min(abs(self.x), abs(self.y), abs(self.z), abs(self.w))

    def maxAbs(self):
        """Return the maximum absolute value of the components.
        """
        return max(abs(self.x), abs(self.y), abs(self.z), abs(self.w))

    def minAbsIndex(self):
        """Return the index of the component with the minimum absolute value.
        """
        ax = abs(self.x)
        ay = abs(self.y)
        az = abs(self.z)
        aw = abs(self.w)
        
        if ax<=ay and ax<=az and ax<=aw:
            return 0
        elif ay<=az and ay<=aw:
            return 1
        elif az<=aw:
            return 2
        else:
            return 3

    def maxAbsIndex(self):
        """Return the index of the component with the maximum absolute value.
        """
        ax = abs(self.x)
        ay = abs(self.y)
        az = abs(self.z)
        aw = abs(self.w)
        
        if ax>=ay and ax>=az and ax>=aw:
            return 0
        elif ay>=az and ay>=aw:
            return 1
        elif az>=aw:
            return 2
        else:
            return 3


    # "t" property (which is an alias for 'w')...
    
    def _getT(self):
        return self.w

    def _setT(self, val):
        self.w = val

    t = property(_getT, _setT, None, "4th component")



######################################################################

def _test():
    import doctest, vec4
    failed, total = doctest.testmod(vec4)
    print(("%d/%d failed" % (failed, total)))

if __name__=="__main__":

    _test()



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the License); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an AS IS basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2010
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the GPL), or
# the GNU Lesser General Public License Version 2.1 or later (the LGPL),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

# Versions this module is based on
LIBAVCODEC_VERSION_INT = 3439104
LIBAVCODEC_VERSION_MAJOR = 52
LIBAVCODEC_VERSION_MICRO = 0
LIBAVCODEC_VERSION_MINOR = 122

LIBAVFORMAT_VERSION_INT = 3436032
LIBAVFORMAT_VERSION_MAJOR = 52
LIBAVFORMAT_VERSION_MICRO = 0
LIBAVFORMAT_VERSION_MINOR = 110

LIBAVUTIL_VERSION_INT = 3287808
LIBAVUTIL_VERSION_MAJOR = 50
LIBAVUTIL_VERSION_MICRO = 0
LIBAVUTIL_VERSION_MINOR = 43

LIBAVFILTER_VERSION_INT = 86016
LIBAVFILTER_VERSION_MAJOR = 1
LIBAVFILTER_VERSION_MICRO = 0
LIBAVFILTER_VERSION_MINOR = 80

LIBSWSCALE_VERSION_INT = 3585
LIBSWSCALE_VERSION_MAJOR = 0
LIBSWSCALE_VERSION_MICRO = 1
LIBSWSCALE_VERSION_MINOR = 14

AV_NOPTS_VALUE = 0x8000000000000000

AV_TIME_BASE = 1000000

MAX_STREAMS = 20

# Error codes
AVERROR_BSF_NOT_FOUND = -1179861752
AVERROR_DECODER_NOT_FOUND = -1128613112
AVERROR_DEMUXER_NOT_FOUND = -1296385272
AVERROR_ENCODER_NOT_FOUND = -1129203192
AVERROR_EOF = -32
AVERROR_EXIT = -1414092869
AVERROR_FILTER_NOT_FOUND = -1279870712
AVERROR_INVALIDDATA = -22
AVERROR_IO = -5
AVERROR_MUXER_NOT_FOUND = -1481985528
AVERROR_NOENT = -2
AVERROR_NOFMT = -92
AVERROR_NOMEM = -12
AVERROR_NOTSUPP = -78
AVERROR_NUMEXPECTED = -33
AVERROR_OPTION_NOT_FOUND = -1414549496
AVERROR_PATCHWELCOME = -1163346256
AVERROR_PROTOCOL_NOT_FOUND = -1330794744
AVERROR_STREAM_NOT_FOUND = -1381258232
AVERROR_UNKNOWN = -22

# CodecType enum values
CODEC_TYPE_ATTACHMENT = 4
CODEC_TYPE_AUDIO = 1
CODEC_TYPE_DATA = 2
CODEC_TYPE_NB = 5
CODEC_TYPE_SUBTITLE = 3
CODEC_TYPE_UNKNOWN = -1
CODEC_TYPE_VIDEO = 0

# Codec flags
CODEC_FLAG_4MV = 4
CODEC_FLAG_AC_PRED = 16777216
CODEC_FLAG_ALT_SCAN = 1048576
CODEC_FLAG_BITEXACT = 8388608
CODEC_FLAG_CBP_RD = 67108864
CODEC_FLAG_CLOSED_GOP = -2147483648
CODEC_FLAG_EMU_EDGE = 16384
CODEC_FLAG_EXTERN_HUFF = 4096
CODEC_FLAG_GLOBAL_HEADER = 4194304
CODEC_FLAG_GMC = 32
CODEC_FLAG_GRAY = 8192
CODEC_FLAG_H263P_AIV = 8
CODEC_FLAG_H263P_SLICE_STRUCT = 268435456
CODEC_FLAG_H263P_UMV = 33554432
CODEC_FLAG_INPUT_PRESERVED = 256
CODEC_FLAG_INTERLACED_DCT = 262144
CODEC_FLAG_INTERLACED_ME = 536870912
CODEC_FLAG_LOOP_FILTER = 2048
CODEC_FLAG_LOW_DELAY = 524288
CODEC_FLAG_MV0 = 64
CODEC_FLAG_NORMALIZE_AQP = 131072
CODEC_FLAG_OBMC = 1
CODEC_FLAG_PART = 128
CODEC_FLAG_PASS1 = 512
CODEC_FLAG_PASS2 = 1024
CODEC_FLAG_PSNR = 32768
CODEC_FLAG_QPEL = 16
CODEC_FLAG_QP_RD = 134217728
CODEC_FLAG_QSCALE = 2
CODEC_FLAG_SVCD_SCAN_OFFSET = 1073741824
CODEC_FLAG_TRUNCATED = 65536

# Flags for the "flags" field in AVOutputFormat/AVInputFormat
AVFMT_FLAG_CUSTOM_IO = 128
AVFMT_FLAG_GENPTS = 1
AVFMT_FLAG_IGNDTS = 8
AVFMT_FLAG_IGNIDX = 2
AVFMT_FLAG_KEEP_SIDE_DATA = 262144
AVFMT_FLAG_MP4A_LATM = 32768
AVFMT_FLAG_NOFILLIN = 16
AVFMT_FLAG_NONBLOCK = 4
AVFMT_FLAG_NOPARSE = 32
AVFMT_FLAG_PRIV_OPT = 131072
AVFMT_FLAG_RTP_HINT = 64
AVFMT_FLAG_SORT_DTS = 65536
AVFMT_GENERIC_INDEX = 256
AVFMT_GLOBALHEADER = 64
AVFMT_INFINITEOUTPUTLOOP = 0
AVFMT_NEEDNUMBER = 2
AVFMT_NOBINSEARCH = 8192
AVFMT_NODIMENSIONS = 2048
AVFMT_NOFILE = 1
AVFMT_NOGENSEARCH = 16384
AVFMT_NOOUTPUTLOOP = -1
AVFMT_NOSTREAMS = 4096
AVFMT_NOTIMESTAMPS = 128
AVFMT_RAWPICTURE = 32
AVFMT_SHOW_IDS = 8
AVFMT_TS_DISCONT = 512
AVFMT_TS_NONSTRICT = 32768
AVFMT_VARIABLE_FPS = 1024

# Flags for url_fopen()
URL_RDONLY = 0
URL_RDWR = 2
URL_WRONLY = 1

# av_seek_frame() flags
AVSEEK_FLAG_ANY = 4
AVSEEK_FLAG_BACKWARD = 1
AVSEEK_FLAG_BYTE = 2
AVSEEK_FLAG_FRAME = 8

# swscale flags
SWS_ACCURATE_RND = 262144
SWS_AREA = 32
SWS_BICUBIC = 4
SWS_BICUBLIN = 64
SWS_BILINEAR = 2
SWS_BITEXACT = 524288
SWS_CPU_CAPS_3DNOW = 1073741824
SWS_CPU_CAPS_ALTIVEC = 268435456
SWS_CPU_CAPS_BFIN = 16777216
SWS_CPU_CAPS_MMX = -2147483648
SWS_CPU_CAPS_MMX2 = 536870912
SWS_CPU_CAPS_SSE2 = 33554432
SWS_CS_DEFAULT = 5
SWS_CS_FCC = 4
SWS_CS_ITU601 = 5
SWS_CS_ITU624 = 5
SWS_CS_ITU709 = 1
SWS_CS_SMPTE170M = 5
SWS_CS_SMPTE240M = 7
SWS_DIRECT_BGR = 32768
SWS_FAST_BILINEAR = 1
SWS_FULL_CHR_H_INP = 16384
SWS_FULL_CHR_H_INT = 8192
SWS_GAUSS = 128
SWS_LANCZOS = 512
SWS_PARAM_DEFAULT = 123456
SWS_POINT = 16
SWS_PRINT_INFO = 4096
SWS_SINC = 256
SWS_SPLINE = 1024
SWS_SRC_V_CHR_DROP_MASK = 196608
SWS_SRC_V_CHR_DROP_SHIFT = 16
SWS_X = 8




# Test job to dump the environment to stdout

import os, socket, getpass
import cgkit.jobqueue

class dumpenv(cgkit.jobqueue.JobProc):
    """Render a Blender file.
    """
    
    def __init__(self):
        cgkit.jobqueue.JobProc.__init__(self, label="Dump Environment")
                 
    def run(self):
        print(("Host name   : %s"%socket.gethostname()))
        print(("User name   : %s"%getpass.getuser()))
        print(("Current dir : %s"%os.getcwd()))
        print ("\nEnvironment variables:\n")
        vars = list(os.environ.keys())
        for var in sorted(vars):
            print(("  %s = %s"%(var, os.environ.get(var))))



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: x3dimport.py,v 1.2 2005/04/15 13:30:35 mbaas Exp $

import os.path
from . import _core
from .cgtypes import *
from .globalscene import getScene
from .worldobject import WorldObject
from .trimesh import TriMesh
from .trimeshgeom import TriMeshGeom
from .targetcamera import TargetCamera
from .glpointlight import GLPointLight
from .glfreespotlight import GLFreeSpotLight
from .glfreedistantlight import GLFreeDistantLight
from .glmaterial import GLMaterial, GLTexture
from .box import Box
from .quadrics import Sphere
from .group import Group
from . import pluginmanager
from .sl import *


# VRMLImporter
class VRMLImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["wrl"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "Virtual Reality Modeling Language (VRML)"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename):
        """Import a VRML file."""

        imp = _core.X3DReader(False)
        imp.read(filename)
        
#        _core.cyber(filename)
#        return

#        sg = _core.X3DSceneGraph()
#        if not sg.load(filename):
#            print "%s (%d): %s"%(sg.getParserErrorMessage(),
#                                 sg.getParserErrorLineNumber(),
#                                 sg.getParserErrorToken())
#            return

#        sg.foo("b = TriMeshGeom()")

# X3DImporter
class X3DImporter:

    _protocols = ["Import"]

    # extension
    def extension():
        """Return the file extensions for this format."""
        return ["x3d"]
    extension = staticmethod(extension)

    # description
    def description(self):
        """Return a short description for the file dialog."""
        return "Extensible 3D (X3D)"
    description = staticmethod(description)

    # importFile
    def importFile(self, filename):
        """Import a X3D file."""

        imp = _core.X3DReader()
        imp.read(filename)

     
######################################################################

# Register the Importer class as a plugin class
if hasattr(_core, "cyber"):
    pluginmanager.register(VRMLImporter)
    pluginmanager.register(X3DImporter)



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2008
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****

"""This module contains the return types used by the slparams and sloargs module.
"""

class _ShaderInfo:
    """Holds information about a shader.
    
    For backwards compatibility this class behaves like a tuple
    (type, name, params). The meta data has to be accessed via attribute
    access.
    """
    
    def __init__(self, type=None, name=None, params=None, meta=None):
        """Constructor.
        """
        if params is None:
            params = []
        if meta is None:
            meta = {}
        
        # The shader type (surface, displacement, ...)
        self.type = type
        # The shader name
        self.name = name
        # The shader parameters (a list of _ShaderParam objects)
        self.params = params
        # The meta data attached to the shader (a dictionary)
        self.meta = meta
        
    def __str__(self):
        return "(%r, %r, %s)"%(self.type, self.name, self.params)
    
    __repr__ = __str__
        
    def __len__(self):
        return 3
    
    def __iter__(self):
        yield self.type
        yield self.name
        yield self.params
        
    def __getitem__(self, idx):
        return (self.type, self.name, self.params)[idx]


class _ShaderParam:
    """Holds information about a shader parameter.
    
    For backwards compatibility this class behaves like a 7-tuple
    (outSpec, storage, type, size, name, space, default). 
    """
    
    def __init__(self, outputSpec=None, storage=None, type=None, size=None, name=None, space=None, default=None):
        """Constructor.
        """
        
        # The output specifier ("output" or empty string)
        self.outputSpec = outputSpec
        # The storage class ("uniform", "varying")
        self.storage = storage
        # The parameter type
        self.type = type
        # The array length or None if the param is not an array
        self.size = size
        # The parameter name
        self.name = name
        # The space (or list of spaces in case of an array) or None
        self.space = space
        # The default value
        self.default = default
        
    def __str__(self):
        return "(%r, %r, %r, %r, %r, %r, %r)"%(self.outputSpec, self.storage, self.type, self.size, self.name, self.space, self.default)
    
    __repr__ = __str__

    def __len__(self):
        return 7
    
    def __iter__(self):
        yield self.outputSpec
        yield self.storage
        yield self.type
        yield self.size
        yield self.name
        yield self.space
        yield self.default
        
    def __getitem__(self, idx):
        return (self.outputSpec, self.storage, self.type, self.size, self.name, self.space, self.default)[idx]

